#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Full Correlation Matrix Analysis (FCMA)

Correlation-based training and prediction
"""

# Authors: Yida Wang
# (Intel Labs), 2016

import numpy as np
import time
from sklearn.base import BaseEstimator
import sklearn
from . import fcma_extension  # type: ignore
from . import cython_blas as blas  # type: ignore
import logging

logger = logging.getLogger(__name__)

__all__ = [
    "Classifier",
]


class Classifier(BaseEstimator):
    """Correlation-based classification component of FCMA

    The classifier first computes correlation of the input data,
    and normalizes them if needed, then uses the given classifier
    to train and/or predict the correlation data.
    NOTE: if the classifier is sklearn.svm.SVC with precomputed kernel,
    the test data may be provided in the fit method to compute
    the kernel matrix together with the training data to save the memory usage,
    but the test data will NEVER be seen in the model training.

    Parameters
    ----------

    clf: class
        The classifier used, normally a classifier class of sklearn

    num_processed_voxels: int, default 2000
        Used for SVM with precomputed kernel,
        every time it only computes correlation between num_process_voxels and
        the whole mask to aggregate the kernel matrices.
        This is to save the memory
        so as to handle correlations at a larger scale.

    epochs_per_subj: int, default 0
        The number of epochs of each subject
        within-subject normalization will be performed during
        classifier training if epochs_per_subj is specified
        default 0 means no within-subject normalization


    Attributes
    ----------

    training_data_: 2D numpy array in shape [num_samples, num_features]
        training_data\\_ is None except clf is SVM.SVC with precomputed kernel,
        in which case training data is needed to compute
        the similarity vector for each sample to be classified.
        However, if the test samples are also provided during the fit,
        the similarity vectors can be precomputed too
        and then training_data\\_ is None

    test_raw_data_: a list of 2D array in shape [num_TRs, num_voxels]
        default None
        test_raw_data\\_ is set after a prediction is called,
        if the new input data equals test_raw_data\\_,
        test_data\\_ can be reused

    test_data_: 2D numpy array in shape [num_samples, num_features]
        default None
        test_data\\_ is set after a prediction is called,
        so that the test data does not need to be regenerated in the
        subsequent operations, e.g. getting decision values of the prediction.
        test_data\\_ may also be set in the fit method
        if sklearn.svm.SVC with precomputed kernel
        and the test samples are known.
        NOTE: the test samples will never be used to fit the model.

    num_voxels_: int
        The number of voxels of the first brain region used in the classifier.
        The first brain region is always large. When training, this region may
        be divided to compute the correlation portion by portion.
        The brain regions are defined by the applied mask, e.g. the top voxels
        selected by FCMA voxel selection

    num_features_: int
        The dimension of correlation data, normally is the product of
        the number of voxels of brain region 1 and
        the number of voxels of brain region 2.
        num_features\\_ must be consistent in both training and classification

    num_samples_: int
        The number of samples

    num_digits_: int
        The number of digits of the first value of the kernel matrix,
        for normalizing the kernel values accordingly
    """
    def __init__(self,
                 clf,
                 num_processed_voxels=2000,
                 epochs_per_subj=0):
        self.clf = clf
        self.num_processed_voxels = num_processed_voxels
        self.epochs_per_subj = epochs_per_subj
        self.num_digits_ = 0
        return

    def _prepare_corerelation_data(self, X1, X2,
                                   start_voxel=0,
                                   num_processed_voxels=None):
        """Compute auto-correlation for the input data X1 and X2.

        it will generate the correlation between some voxels and all voxels

        Parameters
        ----------
        X1: a list of numpy array in shape [num_TRs, num_voxels1]
            X1 contains the activity data filtered by ROIs
            and prepared for correlation computation.
            All elements of X1 must have the same num_voxels value.
        X2: a list of numpy array in shape [num_TRs, num_voxels2]
            len(X1) equals len(X2).
            All elements of X2 must have the same num_voxels value.
            X2 can be identical to X1; if not, X1 must have more voxels
            than X2 (guaranteed by self.fit and/or self.predict).
        start_voxel: int, default 0
            the starting voxel id for correlation computation
        num_processed_voxels: int, default None
            the number of voxels it computes for correlation computation
            if it is None, it is set to self.num_voxels

        Returns
        -------
        corr_data: the correlation data
                    in shape [len(X), num_processed_voxels, num_voxels2]
        """
        num_samples = len(X1)
        assert num_samples > 0, \
            'at least one sample is needed for correlation computation'
        num_voxels1 = X1[0].shape[1]
        num_voxels2 = X2[0].shape[1]
        assert num_voxels1 * num_voxels2 == self.num_features_, \
            'the number of features provided by the input data ' \
            'does not match the number of features defined in the model'
        assert X1[0].shape[0] == X2[0].shape[0], \
            'the numbers of TRs of X1 and X2 are not identical'
        if num_processed_voxels is None:
            num_processed_voxels = num_voxels1
        corr_data = np.zeros((num_samples, num_processed_voxels, num_voxels2),
                             np.float32, order='C')
        # compute correlation
        for idx, data in enumerate(X1):
            data2 = X2[idx]
            num_TRs = data.shape[0]
            blas.compute_corr_vectors('N', 'T',
                                      num_voxels2, num_processed_voxels,
                                      num_TRs,
                                      1.0, data2, num_voxels2,
                                      data, num_voxels1,
                                      0.0, corr_data, num_voxels2,
                                      start_voxel, idx)
        logger.debug(
            'correlation computation done'
        )
        return corr_data

    def _normalize_correlation_data(self, corr_data, norm_unit):
        """Normalize the correlation data if necessary.

        Fisher-transform and then z-score the data for every norm_unit samples
        if norm_unit > 1.

        Parameters
        ----------
        corr_data: the correlation data
                    in shape [num_samples, num_processed_voxels, num_voxels]
        norm_unit: int
                    the number of samples on which the normalization
                    is performed

        Returns
        -------
        normalized_corr_data: the normalized correlation data
                    in shape [num_samples, num_voxels, num_voxels]
        """
        # normalize if necessary
        if norm_unit > 1:
            num_samples = len(corr_data)
            [_, d2, d3] = corr_data.shape
            second_dimension = d2 * d3
            # this is a shallow copy
            normalized_corr_data = corr_data.reshape(1,
                                                     num_samples,
                                                     second_dimension)
            fcma_extension.normalization(normalized_corr_data, norm_unit)
            normalized_corr_data = normalized_corr_data.reshape(num_samples,
                                                                d2, d3)
            logger.debug(
                'normalization done'
            )
        else:
            normalized_corr_data = corr_data
        return normalized_corr_data

    def _prepare_test_data(self, corr_data):
        """Prepare the data to be applied to the predict function.

        if the classifier is SVM, do kernel precomputation,
        otherwise the test data is the reshaped corr_data

        Parameters
        ----------
        corr_data: the (normalized) correlation data
                    in shape [num_samples, num_voxels, num_voxels2]

        Returns
        -------
        data: the data to be predicted, in shape of [num_samples, num_dim]
        """
        num_test_samples = corr_data.shape[0]
        assert num_test_samples > 0, \
            'at least one test sample is needed'
        if isinstance(self.clf, sklearn.svm.SVC) \
                and self.clf.kernel == 'precomputed':
            assert self.training_data_ is not None, \
                'when using precomputed kernel of SVM, ' \
                'all training data must be provided'
            num_training_samples = self.training_data_.shape[0]
            data = np.zeros((num_test_samples, num_training_samples),
                            np.float32,
                            order='C')
            num_features = self.num_features_
            corr_data = corr_data.reshape(num_test_samples,
                                          num_features)
            # compute the similarity vectors using corr_data and training_data
            blas.compute_single_matrix_multiplication('T', 'N',
                                                      num_training_samples,
                                                      num_test_samples,
                                                      num_features,
                                                      1.0,
                                                      self.training_data_,
                                                      num_features,
                                                      corr_data,
                                                      num_features,
                                                      0.0,
                                                      data,
                                                      num_training_samples)
            # shrink the values for getting more stable alpha values
            # in SVM training iteration
            num_digits = self.num_digits_
            if num_digits > 2:
                proportion = 10**(2-num_digits)
                data *= proportion
            logger.debug(
                'similarity vectors computation done'
            )
        else:
            data = corr_data.reshape(num_test_samples,
                                     self.num_features_)
        return data

    def _compute_kernel_matrix_in_portion(self, X1, X2):
        """Compute kernel matrix for sklearn.svm.SVC with precomputed kernel.

        The method generates the kernel matrix (similarity matrix) for
        sklearn.svm.SVC with precomputed kernel. It first computes
        the correlation from X, then normalizes the correlation if needed,
        and finally computes the kernel matrix. It is worth noting that if
        the resulting correlation is large, the kernel matrix will be computed
        portion by portion to save memory usage (the portion size is specified
        in self.num_processed_voxels.

        Parameters
        ----------
        X1: a list of numpy array in shape [num_TRs, num_voxels]
            X1 contains the activity data filtered by ROIs
            and prepared for correlation computation.
            All elements of X1 must have the same num_voxels value,
        X2: a list of numpy array in shape [num_TRs, num_voxels]
            len(X) equals len(X2).
            All elements of X2 must have the same num_voxels value.
            X2 can be identical to X1; if not, X1 always has more voxels
            than X2.

        Returns
        -------
        kernel_matrix: 2D array in shape [num_samples, num_samples]
            the kernel matrix to be used in sklearn.svm.SVC
        normalized_corr_data: 2D array in shape [num_samples, num_features]
            the training data to be used in self.predict() if
            the kernel matrix is computed in one portion,
            otherwise it will not be used.
        """
        kernel_matrix = np.zeros((self.num_samples_, self.num_samples_),
                                 np.float32,
                                 order='C')
        sr = 0
        row_length = self.num_processed_voxels
        num_voxels2 = X2[0].shape[1]
        normalized_corr_data = None
        while sr < self.num_voxels_:
            if row_length >= self.num_voxels_ - sr:
                row_length = self.num_voxels_ - sr
            # compute sub-correlation
            corr_data = self._prepare_corerelation_data(X1, X2,
                                                        sr, row_length)
            # normalization
            normalized_corr_data = self._normalize_correlation_data(
                corr_data,
                self.epochs_per_subj)
            # compute partial kernel matrices
            # for using kernel matrix computation from voxel selection
            normalized_corr_data = normalized_corr_data.reshape(
                1,
                self.num_samples_,
                row_length * num_voxels2)
            blas.compute_kernel_matrix('L', 'T',
                                       self.num_samples_,
                                       row_length * num_voxels2,
                                       1.0, normalized_corr_data,
                                       0, row_length * num_voxels2,
                                       1.0, kernel_matrix, self.num_samples_)
            sr += row_length
        # shrink the values for getting more stable alpha values
        # in SVM training iteration
        num_digits = len(str(int(kernel_matrix[0, 0])))
        self.num_digits_ = num_digits
        if num_digits > 2:
            proportion = 10**(2-num_digits)
            kernel_matrix *= proportion
        return kernel_matrix, normalized_corr_data

    def _generate_training_data(self, X1, X2, num_training_samples):
        """Generate training data for the classifier.

        Compute the correlation, do the normalization if necessary,
        and compute the kernel matrix if the classifier is
        sklearn.svm.SVC with precomputed kernel.

        Parameters
        ----------
        X1: a list of numpy array in shape [num_TRs, num_voxels]
            X1 contains the activity data filtered by ROIs
            and prepared for correlation computation.
            All elements of X1 must have the same num_voxels value,
        X2: a list of numpy array in shape [num_TRs, num_voxels]
            len(X1) equals len(X2).
            All elements of X2 must have the same num_voxels value.
            X2 can be identical to X1; if not, X1 must have more voxels
            than X2 (guaranteed by self.fit).
        num_training_samples: Optional[int]
            Default None.
            The number of samples used in the training,
            which is set when the kernel matrix is constructed
            portion by portion so the similarity vectors of the
            test data have to be computed here.
            This is ONLY set when sklearn.svm.SVC with
            precomputed kernel is used.
            If it is set, only those samples will be used to fit the model.

        Returns
        -------
        data: 2D numpy array
            If the classifier is sklearn.svm.SVC with precomputed kernel,
            data is the kenrl matrix in shape [num_samples, num_samples];
            otherwise, data is in shape [num_samples, num_features] as
            the training data.
        """
        if not (isinstance(self.clf, sklearn.svm.SVC)
                and self.clf.kernel == 'precomputed'):
            # correlation computation
            corr_data = self._prepare_corerelation_data(X1, X2)
            # normalization
            normalized_corr_data = self._normalize_correlation_data(
                corr_data,
                self.epochs_per_subj)
            # training data prepare
            data = normalized_corr_data.reshape(self.num_samples_,
                                                self.num_features_)
            self.training_data_ = None
        else:  # SVM with precomputed kernel
            if self.num_processed_voxels < self.num_voxels_:
                if num_training_samples is None:
                    raise RuntimeError('the kernel matrix will be '
                                       'computed portion by portion, '
                                       'the test samples must be predefined '
                                       'by specifying '
                                       'num_training_samples')
                if num_training_samples >= self.num_samples_:
                    raise ValueError('the number of training samples '
                                     'must be smaller than '
                                     'the number of total samples')
            data, normalized_corr_data = \
                self._compute_kernel_matrix_in_portion(X1, X2)
            if self.num_processed_voxels >= self.num_voxels_:
                # training data is in shape
                # [num_samples, num_voxels * num_voxels]
                self.training_data_ = normalized_corr_data.reshape(
                    self.num_samples_,
                    self.num_features_)
            else:
                # do not store training data because it was partially computed
                self.training_data_ = None
            logger.debug(
                'kernel computation done'
            )
        return data

    def fit(self, X, y, num_training_samples=None):
        """Use correlation data to train a model.

        First compute the correlation of the input data,
        and then normalize within subject
        if more than one sample in one subject,
        and then fit to a model defined by self.clf.

        Parameters
        ----------
        X: list of tuple (data1, data2)
            data1 and data2 are numpy array in shape [num_TRs, num_voxels]
            to be computed for correlation.
            They contain the activity data filtered by ROIs
            and prepared for correlation computation.
            Within list, all data1s must have the same num_voxels value,
            all data2s must have the same num_voxels value.
        y: 1D numpy array
            labels, len(X) equals len(y)
        num_training_samples: Optional[int]
            The number of samples used in the training.
            Set it to construct the kernel matrix
            portion by portion so the similarity vectors of the
            test data have to be computed here.
            Only set num_training_samples when sklearn.svm.SVC with
            precomputed kernel is used.
            If it is set, only those samples will be used to fit the model.

        Returns
        -------
        Classifier:
            self.
        """
        time1 = time.time()
        assert len(X) == len(y), \
            'the number of samples must be equal to the number of labels'
        for x in X:
            assert len(x) == 2, \
                'there must be two parts for each correlation computation'
        X1, X2 = zip(*X)
        if not (isinstance(self.clf, sklearn.svm.SVC)
                and self.clf.kernel == 'precomputed'):
            if num_training_samples is not None:
                num_training_samples = None
                logger.warning(
                    'num_training_samples should not be set for classifiers '
                    'other than SVM with precomputed kernels'
                )
        num_samples = len(X1)
        num_voxels1 = X1[0].shape[1]
        num_voxels2 = X2[0].shape[1]
        # make sure X1 always has more voxels
        if num_voxels1 < num_voxels2:
            X1, X2 = X2, X1
            num_voxels1, num_voxels2 = num_voxels2, num_voxels1
        self.num_voxels_ = num_voxels1
        self.num_features_ = num_voxels1 * num_voxels2
        self.num_samples_ = num_samples

        data = self._generate_training_data(X1, X2, num_training_samples)

        if num_training_samples is not None:
            self.test_raw_data_ = None
            self.test_data_ = data[num_training_samples:,
                                   0:num_training_samples]
            # limit training to the data specified by num_training_samples
            data = data[0:num_training_samples, 0:num_training_samples]
        # training
        self.clf = self.clf.fit(data, y[0:num_training_samples])
        # set the test data
        if num_training_samples is None:
            self.test_raw_data_ = None
            self.test_data_ = None
        time2 = time.time()
        logger.info(
            'training done, takes %.2f s' %
            (time2 - time1)
        )
        return self

    def predict(self, X=None):
        """Use a trained model to predict correlation data.

        first compute the correlation of the input data,
        and then normalize across all samples in the list
        if there are more than one sample,
        and then predict via self.clf.
        If X is None, use the similarity vectors produced in fit
        to predict

        Parameters
        ----------
        X: Optional[list of tuple (data1, data2)]
            data1 and data2 are numpy array in shape [num_TRs, num_voxels]
            to be computed for correlation.
            default None, meaning that the data to be predicted
            have been processed in the fit method.
            Otherwise, X contains the activity data filtered by ROIs
            and prepared for correlation computation.
            len(X) is the number of test samples.
            if len(X) > 1: normalization is done on all test samples.
            Within list, all data1s must have the same num_voxels value,
            all data2s must have the same num_voxels value.

        Returns
        -------
        y_pred: the predicted label of X, in shape [len(X),]
        """
        time1 = time.time()
        if X is not None:
            for x in X:
                assert len(x) == 2, \
                    'there must be two parts for each correlation computation'
            X1, X2 = zip(*X)
            num_voxels1 = X1[0].shape[1]
            num_voxels2 = X2[0].shape[1]
            # make sure X1 always has more voxels
            if num_voxels1 < num_voxels2:
                X1, X2 = X2, X1
                num_voxels1, num_voxels2 = num_voxels2, num_voxels1
            assert self.num_features_ == num_voxels1 * num_voxels2, \
                'the number of features does not match the model'
            num_test_samples = len(X1)
            self.test_raw_data_ = X

            # correlation computation
            corr_data = self._prepare_corerelation_data(X1, X2)
            # normalization
            normalized_corr_data = self._normalize_correlation_data(
                corr_data,
                num_test_samples)
            # test data generation
            self.test_data_ = self._prepare_test_data(normalized_corr_data)
        # prediction
        y_pred = self.clf.predict(self.test_data_)
        time2 = time.time()
        logger.info(
            'prediction done, takes %.2f s' %
            (time2 - time1)
        )
        return y_pred

    def _is_equal_to_test_raw_data(self, X):
        """Check if the new input data X is equal to the old one.

        compare X and self.test_raw_data_ if it exists

        Parameters
        ----------
        X: list of tuple (data1, data2)
            data1 and data2 are numpy array in shape [num_TRs, num_voxels],
            the input data to be checked.

        Returns
        -------
        a boolean value to indicate if X == self.test_raw_data_
        """
        if self.test_raw_data_ is None or len(X) != len(self.test_raw_data_):
            return False
        X1, X2 = zip(*X)
        c1, c2 = zip(*self.test_raw_data_)
        # this for loop is faster than
        # doing np.array_equal(X, self.test_raw_data_) directly
        for new, old in zip(X1, c1):
            if not np.array_equal(new, old):
                return False
        for new, old in zip(X2, c2):
            if not np.array_equal(new, old):
                return False
        return True

    def decision_function(self, X=None):
        """Output the decision value of the prediction.

        if X is not equal to self.test_raw_data\\_, i.e. predict is not called,
        first generate the test_data
        after getting the test_data, get the decision value via self.clf.
        if X is None, test_data\\_ is ready to be used

        Parameters
        ----------
        X: Optional[list of tuple (data1, data2)]
            data1 and data2 are numpy array in shape [num_TRs, num_voxels]
            to be computed for correlation.
            default None, meaning that the data to be predicted
            have been processed in the fit method.
            Otherwise, X contains the activity data filtered by ROIs
            and prepared for correlation computation.
            len(X) is the number of test samples.
            if len(X) > 1: normalization is done on all test samples.
            Within list, all data1s must have the same num_voxels value,
            all data2s must have the same num_voxels value.

        Returns
        -------
        confidence: the predictions confidence values of X, in shape [len(X),]
        """
        if X is not None and not self._is_equal_to_test_raw_data(X):
            for x in X:
                assert len(x) == 2, \
                    'there must be two parts for each correlation computation'
            X1, X2 = zip(*X)
            num_voxels1 = X1[0].shape[1]
            num_voxels2 = X2[0].shape[1]
            assert len(X1) == len(X2), \
                'the list lengths do not match'
            # make sure X1 always has more voxels
            if num_voxels1 < num_voxels2:
                X1, X2 = X2, X1
                num_voxels1, num_voxels2 = num_voxels2, num_voxels1
            assert self.num_features_ == num_voxels1 * num_voxels2, \
                'the number of features does not match the model'
            num_test_samples = len(X1)
            self.test_raw_data_ = X
            # generate the test_data first
            # correlation computation
            corr_data = self._prepare_corerelation_data(X1, X2)
            # normalization
            normalized_corr_data = \
                self._normalize_correlation_data(corr_data,
                                                 num_test_samples)
            # test data generation
            self.test_data_ = self._prepare_test_data(normalized_corr_data)
        confidence = self.clf.decision_function(self.test_data_)
        return confidence

    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.

        NOTE: In the condition of sklearn.svm.SVC with precomputed kernel
        when the kernel matrix is computed portion by portion, the function
        will ignore the first input argument X.

        Parameters
        ----------
        X: list of tuple (data1, data2)
            data1 and data2 are numpy array in shape [num_TRs, num_voxels]
            to be computed for correlation.
            They are test samples.
            They contain the activity data filtered by ROIs
            and prepared for correlation computation.
            Within list, all data1s must have the same num_voxels value,
            all data2s must have the same num_voxels value.
            len(X) is the number of test samples.

        y: 1D numpy array
            labels, len(X) equals len(y), which is num_samples
        sample_weight: 1D array in shape [num_samples], optional
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) wrt. y.
        """
        from sklearn.metrics import accuracy_score
        if isinstance(self.clf, sklearn.svm.SVC) \
                and self.clf.kernel == 'precomputed' \
                and self.training_data_ is None:
            result = accuracy_score(y, self.predict(),
                                    sample_weight=sample_weight)
        else:
            result = accuracy_score(y, self.predict(X),
                                    sample_weight=sample_weight)
        return result



import numpy as np
import logging
import pytest
from brainiak.isc import (isc, isfc, bootstrap_isc, permutation_isc,
                          squareform_isfc, timeshift_isc,
                          phaseshift_isc)
from scipy.spatial.distance import squareform

logger = logging.getLogger(__name__)


# Create simple simulated data with high intersubject correlation
def simulated_timeseries(n_subjects, n_TRs, n_voxels=30,
                         noise=1, data_type='array',
                         random_state=None):
    prng = np.random.RandomState(random_state)
    if n_voxels:
        signal = prng.randn(n_TRs, n_voxels)
        prng = np.random.RandomState(prng.randint(0, 2**32 - 1,
                                                  dtype=np.int64))
        data = [signal + prng.randn(n_TRs, n_voxels) * noise
                for subject in np.arange(n_subjects)]
    elif not n_voxels:
        signal = prng.randn(n_TRs)
        prng = np.random.RandomState(prng.randint(0, 2**32 - 1,
                                                  dtype=np.int64))
        data = [signal + prng.randn(n_TRs) * noise
                for subject in np.arange(n_subjects)]
    if data_type == 'array':
        if n_voxels:
            data = np.dstack(data)
        elif not n_voxels:
            data = np.column_stack(data)
    return data


# Create 3 voxel simulated data with correlated time series
def correlated_timeseries(n_subjects, n_TRs, noise=0,
                          random_state=None):
    prng = np.random.RandomState(random_state)
    signal = prng.randn(n_TRs)
    correlated = True
    while correlated:
        uncorrelated = np.random.randn(n_TRs,
                                       n_subjects)[:, np.newaxis, :]
        unc_max = np.amax(squareform(np.corrcoef(
            uncorrelated[:, 0, :].T), checks=False))
        unc_mean = np.mean(squareform(np.corrcoef(
            uncorrelated[:, 0, :].T), checks=False))
        if unc_max < .25 and np.abs(unc_mean) < .001:
            correlated = False
    data = np.repeat(np.column_stack((signal, signal))[..., np.newaxis],
                     n_subjects, axis=2)
    data = np.concatenate((data, uncorrelated), axis=1)
    data = data + np.random.randn(n_TRs, 3, n_subjects) * noise
    return data


# Compute ISCs using different input types
# List of subjects with one voxel/ROI
def test_isc_input():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42

    logger.info("Testing ISC inputs")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=None, data_type='list',
                                random_state=random_state)
    iscs_list = isc(data, pairwise=False, summary_statistic=None)

    # Array of subjects with one voxel/ROI
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=None, data_type='array',
                                random_state=random_state)
    iscs_array = isc(data, pairwise=False, summary_statistic=None)

    # Check they're the same
    assert np.array_equal(iscs_list, iscs_array)

    # List of subjects with multiple voxels/ROIs
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='list',
                                random_state=random_state)
    iscs_list = isc(data, pairwise=False, summary_statistic=None)

    # Array of subjects with multiple voxels/ROIs
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)
    iscs_array = isc(data, pairwise=False, summary_statistic=None)

    # Check they're the same
    assert np.array_equal(iscs_list, iscs_array)

    logger.info("Finished testing ISC inputs")


# Check pairwise and leave-one-out, and summary statistics for ISC
def test_isc_options():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42

    logger.info("Testing ISC options")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    iscs_loo = isc(data, pairwise=False, summary_statistic=None)
    assert iscs_loo.shape == (n_subjects, n_voxels)

    # Just two subjects
    iscs_loo = isc(data[..., :2], pairwise=False, summary_statistic=None)
    assert iscs_loo.shape == (n_voxels,)

    iscs_pw = isc(data, pairwise=True, summary_statistic=None)
    assert iscs_pw.shape == (n_subjects*(n_subjects-1)/2, n_voxels)

    # Check summary statistics
    isc_mean = isc(data, pairwise=False, summary_statistic='mean')
    assert isc_mean.shape == (n_voxels,)

    isc_median = isc(data, pairwise=False, summary_statistic='median')
    assert isc_median.shape == (n_voxels,)

    with pytest.raises(ValueError):
        isc(data, pairwise=False, summary_statistic='min')

    logger.info("Finished testing ISC options")


# Make sure ISC recovers correlations of 1 and less than 1
def test_isc_output():

    logger.info("Testing ISC outputs")

    data = correlated_timeseries(20, 60, noise=0,
                                 random_state=42)
    iscs = isc(data, pairwise=False)
    assert np.allclose(iscs[:, :2], 1., rtol=1e-05)
    assert np.all(iscs[:, -1] < 1.)

    iscs = isc(data, pairwise=True)
    assert np.allclose(iscs[:, :2], 1., rtol=1e-05)
    assert np.all(iscs[:, -1] < 1.)

    logger.info("Finished testing ISC outputs")


# Check for proper handling of NaNs in ISC
def test_isc_nans():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42

    logger.info("Testing ISC options")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    # Inject NaNs into data
    data[0, 0, 0] = np.nan

    # Don't tolerate NaNs, should lose zeroeth voxel
    iscs_loo = isc(data, pairwise=False, tolerate_nans=False)
    assert np.sum(np.isnan(iscs_loo)) == n_subjects

    # Tolerate all NaNs, only subject with NaNs yields NaN
    iscs_loo = isc(data, pairwise=False, tolerate_nans=True)
    assert np.sum(np.isnan(iscs_loo)) == 1

    # Pairwise approach shouldn't care
    iscs_pw_T = isc(data, pairwise=True, tolerate_nans=True)
    iscs_pw_F = isc(data, pairwise=True, tolerate_nans=False)
    assert np.allclose(iscs_pw_T, iscs_pw_F, equal_nan=True)

    assert (np.sum(np.isnan(iscs_pw_T)) ==
            np.sum(np.isnan(iscs_pw_F)) ==
            n_subjects - 1)

    # Set proportion of nans to reject (70% and 90% non-NaN)
    data[0, 0, :] = np.nan
    data[0, 1, :n_subjects - int(n_subjects * .7)] = np.nan
    data[0, 2, :n_subjects - int(n_subjects * .9)] = np.nan

    iscs_loo_T = isc(data, pairwise=False, tolerate_nans=True)
    iscs_loo_F = isc(data, pairwise=False, tolerate_nans=False)
    iscs_loo_95 = isc(data, pairwise=False, tolerate_nans=.95)
    iscs_loo_90 = isc(data, pairwise=False, tolerate_nans=.90)
    iscs_loo_80 = isc(data, pairwise=False, tolerate_nans=.8)
    iscs_loo_70 = isc(data, pairwise=False, tolerate_nans=.7)
    iscs_loo_60 = isc(data, pairwise=False, tolerate_nans=.6)

    assert (np.sum(np.isnan(iscs_loo_F)) ==
            np.sum(np.isnan(iscs_loo_95)) == 60)
    assert (np.sum(np.isnan(iscs_loo_80)) ==
            np.sum(np.isnan(iscs_loo_90)) == 42)
    assert (np.sum(np.isnan(iscs_loo_T)) ==
            np.sum(np.isnan(iscs_loo_60)) ==
            np.sum(np.isnan(iscs_loo_70)) == 28)
    assert np.array_equal(np.sum(np.isnan(iscs_loo_F), axis=0),
                          np.sum(np.isnan(iscs_loo_95), axis=0))
    assert np.array_equal(np.sum(np.isnan(iscs_loo_80), axis=0),
                          np.sum(np.isnan(iscs_loo_90), axis=0))
    assert np.all((np.array_equal(
                        np.sum(np.isnan(iscs_loo_T), axis=0),
                        np.sum(np.isnan(iscs_loo_60), axis=0)),
                   np.array_equal(
                        np.sum(np.isnan(iscs_loo_T), axis=0),
                        np.sum(np.isnan(iscs_loo_70), axis=0)),
                   np.array_equal(
                        np.sum(np.isnan(iscs_loo_60), axis=0),
                        np.sum(np.isnan(iscs_loo_70), axis=0))))

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    # Make sure voxel with NaNs across all subjects is always removed
    data[0, 0, :] = np.nan
    iscs_loo_T = isc(data, pairwise=False, tolerate_nans=True)
    iscs_loo_F = isc(data, pairwise=False, tolerate_nans=False)
    assert np.allclose(iscs_loo_T, iscs_loo_F, equal_nan=True)
    assert (np.sum(np.isnan(iscs_loo_T)) ==
            np.sum(np.isnan(iscs_loo_F)) ==
            n_subjects)

    iscs_pw_T = isc(data, pairwise=True, tolerate_nans=True)
    iscs_pw_F = isc(data, pairwise=True, tolerate_nans=False)
    assert np.allclose(iscs_pw_T, iscs_pw_F, equal_nan=True)

    assert (np.sum(np.isnan(iscs_pw_T)) ==
            np.sum(np.isnan(iscs_pw_F)) ==
            n_subjects * (n_subjects - 1) / 2)


# Test one-sample bootstrap test
def test_bootstrap_isc():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42
    n_bootstraps = 10

    logger.info("Testing bootstrap hypothesis test")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    iscs = isc(data, pairwise=False, summary_statistic=None)
    observed, ci, p, distribution = bootstrap_isc(iscs, pairwise=False,
                                                  summary_statistic='median',
                                                  n_bootstraps=n_bootstraps,
                                                  ci_percentile=95)
    assert distribution.shape == (n_bootstraps, n_voxels)

    # Test one-sample bootstrap test with pairwise approach
    n_bootstraps = 10

    iscs = isc(data, pairwise=True, summary_statistic=None)
    observed, ci, p, distribution = bootstrap_isc(iscs, pairwise=True,
                                                  summary_statistic='median',
                                                  n_bootstraps=n_bootstraps,
                                                  ci_percentile=95)
    assert distribution.shape == (n_bootstraps, n_voxels)

    # Check random seeds
    iscs = isc(data, pairwise=False, summary_statistic=None)
    distributions = []
    for random_state in [42, 42, None]:
        observed, ci, p, distribution = bootstrap_isc(
                                                iscs, pairwise=False,
                                                summary_statistic='median',
                                                n_bootstraps=n_bootstraps,
                                                ci_percentile=95,
                                                random_state=random_state)
        distributions.append(distribution)
    assert np.array_equal(distributions[0], distributions[1])
    assert not np.array_equal(distributions[1], distributions[2])

    # Check output p-values
    data = correlated_timeseries(20, 60, noise=.5,
                                 random_state=42)
    iscs = isc(data, pairwise=False)
    observed, ci, p, distribution = bootstrap_isc(iscs, pairwise=False)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01
    print(p)

    iscs = isc(data, pairwise=True)
    observed, ci, p, distribution = bootstrap_isc(iscs, pairwise=True)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    # Check that ISC computation and bootstrap observed are same
    iscs = isc(data, pairwise=False)
    observed, ci, p, distribution = bootstrap_isc(iscs, pairwise=False,
                                                  summary_statistic='median')
    assert np.array_equal(observed, isc(data, pairwise=False,
                                        summary_statistic='median'))

    # Check that ISC computation and bootstrap observed are same
    iscs = isc(data, pairwise=True)
    observed, ci, p, distribution = bootstrap_isc(iscs, pairwise=True,
                                                  summary_statistic='median')
    assert np.array_equal(observed, isc(data, pairwise=True,
                                        summary_statistic='median'))

    logger.info("Finished testing bootstrap hypothesis test")


# Test permutation test with group assignments
def test_permutation_isc():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42
    group_assignment = [1] * 10 + [2] * 10

    logger.info("Testing permutation test")

    # Create dataset with two groups in pairwise approach
    data = np.dstack((simulated_timeseries(10, n_TRs, n_voxels=n_voxels,
                                           noise=1, data_type='array',
                                           random_state=3),
                      simulated_timeseries(10, n_TRs, n_voxels=n_voxels,
                                           noise=5, data_type='array',
                                           random_state=4)))
    iscs = isc(data, pairwise=True, summary_statistic=None)

    observed, p, distribution = permutation_isc(
                                            iscs,
                                            group_assignment=group_assignment,
                                            pairwise=True,
                                            summary_statistic='mean',
                                            n_permutations=200)

    # Create data with two groups in leave-one-out approach
    data_1 = simulated_timeseries(10, n_TRs, n_voxels=n_voxels,
                                  noise=1, data_type='array',
                                  random_state=3)
    data_2 = simulated_timeseries(10, n_TRs, n_voxels=n_voxels,
                                  noise=10, data_type='array',
                                  random_state=4)
    iscs = np.vstack((isc(data_1, pairwise=False, summary_statistic=None),
                      isc(data_2, pairwise=False, summary_statistic=None)))

    observed, p, distribution = permutation_isc(
                                            iscs,
                                            group_assignment=group_assignment,
                                            pairwise=False,
                                            summary_statistic='mean',
                                            n_permutations=200)

    # One-sample leave-one-out permutation test
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)
    iscs = isc(data, pairwise=False, summary_statistic=None)

    observed, p, distribution = permutation_isc(iscs,
                                                pairwise=False,
                                                summary_statistic='median',
                                                n_permutations=200)

    # One-sample pairwise permutation test
    iscs = isc(data, pairwise=True, summary_statistic=None)

    observed, p, distribution = permutation_isc(iscs,
                                                pairwise=True,
                                                summary_statistic='median',
                                                n_permutations=200)

    # Small one-sample pairwise exact test
    data = simulated_timeseries(12, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)
    iscs = isc(data, pairwise=False, summary_statistic=None)

    observed, p, distribution = permutation_isc(iscs, pairwise=False,
                                                summary_statistic='median',
                                                n_permutations=10000)

    # Small two-sample pairwise exact test (and unequal groups)
    data = np.dstack((simulated_timeseries(3, n_TRs, n_voxels=n_voxels,
                                           noise=1, data_type='array',
                                           random_state=3),
                      simulated_timeseries(4, n_TRs, n_voxels=n_voxels,
                                           noise=50, data_type='array',
                                           random_state=4)))
    iscs = isc(data, pairwise=True, summary_statistic=None)
    group_assignment = [1, 1, 1, 2, 2, 2, 2]

    observed, p, distribution = permutation_isc(
                                            iscs,
                                            group_assignment=group_assignment,
                                            pairwise=True,
                                            summary_statistic='mean',
                                            n_permutations=10000)

    # Small two-sample leave-one-out exact test (and unequal groups)
    data_1 = simulated_timeseries(3, n_TRs, n_voxels=n_voxels,
                                  noise=1, data_type='array',
                                  random_state=3)
    data_2 = simulated_timeseries(4, n_TRs, n_voxels=n_voxels,
                                  noise=50, data_type='array',
                                  random_state=4)
    iscs = np.vstack((isc(data_1, pairwise=False, summary_statistic=None),
                      isc(data_2, pairwise=False, summary_statistic=None)))
    group_assignment = [1, 1, 1, 2, 2, 2, 2]

    observed, p, distribution = permutation_isc(
                                            iscs,
                                            group_assignment=group_assignment,
                                            pairwise=False,
                                            summary_statistic='mean',
                                            n_permutations=10000)

    # Check output p-values
    data = correlated_timeseries(20, 60, noise=.5,
                                 random_state=42)
    iscs = isc(data, pairwise=False)
    observed, p, distribution = permutation_isc(iscs, pairwise=False)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    iscs = isc(data, pairwise=True)
    observed, p, distribution = permutation_isc(iscs, pairwise=True)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    # Check that ISC computation and permutation observed are same
    iscs = isc(data, pairwise=False)
    observed, p, distribution = permutation_isc(iscs, pairwise=False,
                                                summary_statistic='median')
    assert np.allclose(observed, isc(data, pairwise=False,
                                     summary_statistic='median'),
                       rtol=1e-03)

    # Check that ISC computation and permuation observed are same
    iscs = isc(data, pairwise=True)
    observed, p, distribution = permutation_isc(iscs, pairwise=True,
                                                summary_statistic='mean')
    assert np.allclose(observed, isc(data, pairwise=True,
                                     summary_statistic='mean'),
                       rtol=1e-03)

    logger.info("Finished testing permutaton test")


def test_timeshift_isc():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30

    logger.info("Testing circular time-shift")

    # Circular time-shift on one sample, leave-one-out
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    observed, p, distribution = timeshift_isc(data, pairwise=False,
                                              summary_statistic='median',
                                              n_shifts=200)

    # Circular time-shift on one sample, pairwise
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    observed, p, distribution = timeshift_isc(data, pairwise=True,
                                              summary_statistic='median',
                                              n_shifts=200)

    # Circular time-shift on one sample, leave-one-out
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    observed, p, distribution = timeshift_isc(data, pairwise=False,
                                              summary_statistic='mean',
                                              n_shifts=200)
    # Check output p-values
    data = correlated_timeseries(20, 60, noise=.5,
                                 random_state=42)
    iscs = isc(data, pairwise=False)
    observed, p, distribution = timeshift_isc(data, pairwise=False)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    iscs = isc(data, pairwise=True)
    observed, p, distribution = timeshift_isc(data, pairwise=True)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    # Check that ISC computation and permutation observed are same
    iscs = isc(data, pairwise=False)
    observed, p, distribution = timeshift_isc(data, pairwise=False,
                                              summary_statistic='median')
    assert np.allclose(observed, isc(data, pairwise=False,
                                     summary_statistic='median'),
                       rtol=1e-03)

    # Check that ISC computation and permuation observed are same
    iscs = isc(data, pairwise=True)
    observed, p, distribution = timeshift_isc(data, pairwise=True,
                                              summary_statistic='mean')
    assert np.allclose(observed, isc(data, pairwise=True,
                                     summary_statistic='mean'),
                       rtol=1e-03)

    logger.info("Finished testing circular time-shift")


# Phase randomization test
def test_phaseshift_isc():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30

    logger.info("Testing phase randomization")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    observed, p, distribution = phaseshift_isc(data, pairwise=True,
                                               summary_statistic='median',
                                               n_shifts=200)

    # Phase randomization one-sample test, leave-one-out
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    observed, p, distribution = phaseshift_isc(data, pairwise=False,
                                               summary_statistic='mean',
                                               n_shifts=200)

    # Check output p-values
    data = correlated_timeseries(20, 60, noise=.5,
                                 random_state=42)
    iscs = isc(data, pairwise=False)
    observed, p, distribution = phaseshift_isc(data, pairwise=False)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    iscs = isc(data, pairwise=True)
    observed, p, distribution = phaseshift_isc(data, pairwise=True)
    assert np.all(iscs[:, :2] > .5)
    assert np.all(iscs[:, -1] < .5)
    assert p[0] < .05 and p[1] < .05
    assert p[2] > .01

    # Check that ISC computation and permutation observed are same
    iscs = isc(data, pairwise=False)
    observed, p, distribution = phaseshift_isc(data, pairwise=False,
                                               summary_statistic='median')
    assert np.allclose(observed, isc(data, pairwise=False,
                                     summary_statistic='median'),
                       rtol=1e-03)

    # Check that ISC computation and permuation observed are same
    iscs = isc(data, pairwise=True)
    observed, p, distribution = phaseshift_isc(data, pairwise=True,
                                               summary_statistic='mean')
    assert np.allclose(observed, isc(data, pairwise=True,
                                     summary_statistic='mean'),
                       rtol=1e-03)

    logger.info("Finished testing phase randomization")


# Test ISFC
def test_isfc_options():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30

    logger.info("Testing ISFC options")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    isfcs, iscs = isfc(data, pairwise=False, summary_statistic=None)
    assert isfcs.shape == (n_subjects, n_voxels * (n_voxels - 1) / 2)
    assert iscs.shape == (n_subjects, n_voxels)

    # Without vectorized upper triangle
    isfcs = isfc(data, pairwise=False, summary_statistic=None,
                 vectorize_isfcs=False)
    assert isfcs.shape == (n_subjects, n_voxels, n_voxels)

    # Just two subjects
    isfcs, iscs = isfc(data[..., :2], pairwise=False, summary_statistic=None)
    assert isfcs.shape == (n_voxels * (n_voxels - 1) / 2,)
    assert iscs.shape == (n_voxels,)

    isfcs = isfc(data[..., :2], pairwise=False, summary_statistic=None,
                 vectorize_isfcs=False)
    assert isfcs.shape == (n_voxels, n_voxels)

    # ISFC with pairwise approach
    isfcs, iscs = isfc(data, pairwise=True, summary_statistic=None)
    assert isfcs.shape == (n_subjects * (n_subjects - 1) / 2,
                           n_voxels * (n_voxels - 1) / 2)
    assert iscs.shape == (n_subjects * (n_subjects - 1) / 2,
                          n_voxels)

    isfcs = isfc(data, pairwise=True, summary_statistic=None,
                 vectorize_isfcs=False)
    assert isfcs.shape == (n_subjects * (n_subjects - 1) / 2,
                           n_voxels, n_voxels)

    # ISFC with summary statistics
    isfcs, iscs = isfc(data, pairwise=True, summary_statistic='mean')
    isfcs, iscs = isfc(data, pairwise=True, summary_statistic='median')

    # Check output p-values
    data = correlated_timeseries(20, 60, noise=.5,
                                 random_state=42)
    isfcs = isfc(data, pairwise=False, vectorize_isfcs=False)
    assert np.all(isfcs[:, 0, 1] > .5) and np.all(isfcs[:, 1, 0] > .5)
    assert np.all(isfcs[:, :2, 2] < .5) and np.all(isfcs[:, 2, :2] < .5)

    isfcs = isfc(data, pairwise=True, vectorize_isfcs=False)
    assert np.all(isfcs[:, 0, 1] > .5) and np.all(isfcs[:, 1, 0] > .5)
    assert np.all(isfcs[:, :2, 2] < .5) and np.all(isfcs[:, 2, :2] < .5)

    # Check that ISC and ISFC diagonal are identical
    iscs = isc(data, pairwise=False)
    isfcs = isfc(data, pairwise=False, vectorize_isfcs=False)
    for s in np.arange(len(iscs)):
        assert np.allclose(isfcs[s, ...].diagonal(), iscs[s, :], rtol=1e-03)
    isfcs, iscs_v = isfc(data, pairwise=False)
    assert np.allclose(iscs, iscs_v, rtol=1e-03)

    # Check that ISC and ISFC diagonal are identical (pairwise)
    iscs = isc(data, pairwise=True)
    isfcs = isfc(data, pairwise=True, vectorize_isfcs=False)
    for s in np.arange(len(iscs)):
        assert np.allclose(isfcs[s, ...].diagonal(), iscs[s, :], rtol=1e-03)
    isfcs, iscs_v = isfc(data, pairwise=True)
    assert np.allclose(iscs, iscs_v, rtol=1e-03)

    # Generate 'targets' data and use for ISFC
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array')
    n_targets = 15
    targets_data = simulated_timeseries(n_subjects, n_TRs,
                                        n_voxels=n_targets,
                                        data_type='array')
    isfcs = isfc(data, targets=targets_data, pairwise=False,
                 vectorize_isfcs=False)
    assert isfcs.shape == (n_subjects, n_voxels, n_targets)

    # Ensure 'square' output enforced
    isfcs = isfc(data, targets=targets_data, pairwise=False,
                 vectorize_isfcs=True)
    assert isfcs.shape == (n_subjects, n_voxels, n_targets)

    # Check list input for targets
    targets_data = simulated_timeseries(n_subjects, n_TRs,
                                        n_voxels=n_targets,
                                        data_type='list')
    isfcs = isfc(data, targets=targets_data, pairwise=False,
                 vectorize_isfcs=False)
    assert isfcs.shape == (n_subjects, n_voxels, n_targets)

    # Check that mismatching subjects / TRs breaks targets
    targets_data = simulated_timeseries(n_subjects, n_TRs,
                                        n_voxels=n_targets,
                                        data_type='array')

    with pytest.raises(ValueError):
        isfcs = isfc(data, targets=targets_data[..., :-1],
                     pairwise=False, vectorize_isfcs=False)
    assert isfcs.shape == (n_subjects, n_voxels, n_targets)

    with pytest.raises(ValueError):
        isfcs = isfc(data, targets=targets_data[:-1, ...],
                     pairwise=False, vectorize_isfcs=False)

    # Check targets for only 2 subjects
    isfcs = isfc(data[..., :2], targets=targets_data[..., :2],
                 pairwise=False, summary_statistic=None)
    assert isfcs.shape == (2, n_voxels, n_targets)

    isfcs = isfc(data[..., :2], targets=targets_data[..., :2],
                 pairwise=True, summary_statistic=None)
    assert isfcs.shape == (2, n_voxels, n_targets)

    # Check that supplying targets enforces leave-one-out
    isfcs_pw = isfc(data, targets=targets_data, pairwise=True,
                    vectorize_isfcs=False, tolerate_nans=False)
    assert isfcs_pw.shape == (n_subjects, n_voxels, n_targets)

    logger.info("Finished testing ISFC options")


# Check for proper handling of NaNs in ISFC
def test_isfc_nans():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42

    logger.info("Testing ISC options")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    # Inject NaNs into data
    data[0, 0, 0] = np.nan

    # Don't tolerate NaNs, should lose zeroeth voxel
    isfcs_loo = isfc(data, pairwise=False, vectorize_isfcs=False,
                     tolerate_nans=False)
    assert np.sum(np.isnan(isfcs_loo)) == n_subjects * (n_voxels * 2 - 1)

    # With vectorized ISFCs
    isfcs_loo, iscs_loo = isfc(data, pairwise=False, vectorize_isfcs=True,
                               tolerate_nans=False)
    assert np.sum(np.isnan(isfcs_loo)) == n_subjects * (n_voxels - 1)

    # Tolerate all NaNs, only subject with NaNs yields NaN
    isfcs_loo = isfc(data, pairwise=False, vectorize_isfcs=False,
                     tolerate_nans=True)
    assert np.sum(np.isnan(isfcs_loo)) == n_voxels * 2 - 1

    isfcs_loo, iscs_loo = isfc(data, pairwise=False, vectorize_isfcs=True,
                               tolerate_nans=True)
    assert np.sum(np.isnan(isfcs_loo)) == n_voxels - 1

    # Pairwise approach shouldn't care
    isfcs_pw_T = isfc(data, pairwise=True, vectorize_isfcs=False,
                      tolerate_nans=True)
    isfcs_pw_F = isfc(data, pairwise=True, vectorize_isfcs=False,
                      tolerate_nans=False)
    assert np.allclose(isfcs_pw_T, isfcs_pw_F, equal_nan=True)
    assert (np.sum(np.isnan(isfcs_pw_T)) ==
            np.sum(np.isnan(isfcs_pw_F)) ==
            (n_voxels * 2 - 1) * (n_subjects - 1))

    isfcs_pw_T, iscs_pw_T = isfc(data, pairwise=True, vectorize_isfcs=True,
                                 tolerate_nans=True)
    isfcs_pw_F, iscs_pw_T = isfc(data, pairwise=True, vectorize_isfcs=True,
                                 tolerate_nans=False)
    assert np.allclose(isfcs_pw_T, isfcs_pw_F, equal_nan=True)
    assert (np.sum(np.isnan(isfcs_pw_T)) ==
            np.sum(np.isnan(isfcs_pw_F)) ==
            (n_voxels - 1) * (n_subjects - 1))

    # Set proportion of nans to reject (70% and 90% non-NaN)
    data[0, 0, :] = np.nan
    data[0, 1, :n_subjects - int(n_subjects * .7)] = np.nan
    data[0, 2, :n_subjects - int(n_subjects * .9)] = np.nan

    isfcs_loo_T = isfc(data, pairwise=False, vectorize_isfcs=False,
                       tolerate_nans=True)
    isfcs_loo_F = isfc(data, pairwise=False, vectorize_isfcs=False,
                       tolerate_nans=False)
    isfcs_loo_95 = isfc(data, pairwise=False, vectorize_isfcs=False,
                        tolerate_nans=.95)
    isfcs_loo_90 = isfc(data, pairwise=False, vectorize_isfcs=False,
                        tolerate_nans=.90)
    isfcs_loo_80 = isfc(data, pairwise=False, vectorize_isfcs=False,
                        tolerate_nans=.8)
    isfcs_loo_70 = isfc(data, pairwise=False, vectorize_isfcs=False,
                        tolerate_nans=.7)
    isfcs_loo_60 = isfc(data, pairwise=False, vectorize_isfcs=False,
                        tolerate_nans=.6)
    assert (np.sum(np.isnan(isfcs_loo_F)) ==
            np.sum(np.isnan(isfcs_loo_95)) == 3420)
    assert (np.sum(np.isnan(isfcs_loo_80)) ==
            np.sum(np.isnan(isfcs_loo_90)) == 2430)
    assert (np.sum(np.isnan(isfcs_loo_T)) ==
            np.sum(np.isnan(isfcs_loo_60)) ==
            np.sum(np.isnan(isfcs_loo_70)) == 1632)
    assert np.array_equal(np.sum(np.isnan(isfcs_loo_F), axis=0),
                          np.sum(np.isnan(isfcs_loo_95), axis=0))
    assert np.array_equal(np.sum(np.isnan(isfcs_loo_80), axis=0),
                          np.sum(np.isnan(isfcs_loo_90), axis=0))
    assert np.all((np.array_equal(
                        np.sum(np.isnan(isfcs_loo_T), axis=0),
                        np.sum(np.isnan(isfcs_loo_60), axis=0)),
                   np.array_equal(
                        np.sum(np.isnan(isfcs_loo_T), axis=0),
                        np.sum(np.isnan(isfcs_loo_70), axis=0)),
                   np.array_equal(
                        np.sum(np.isnan(isfcs_loo_60), axis=0),
                        np.sum(np.isnan(isfcs_loo_70), axis=0))))

    isfcs_loo_T, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                          tolerate_nans=True)
    isfcs_loo_F, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                          tolerate_nans=False)
    isfcs_loo_95, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                           tolerate_nans=.95)
    isfcs_loo_90, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                           tolerate_nans=.90)
    isfcs_loo_80, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                           tolerate_nans=.8)
    isfcs_loo_70, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                           tolerate_nans=.7)
    isfcs_loo_60, _ = isfc(data, pairwise=False, vectorize_isfcs=True,
                           tolerate_nans=.6)
    assert (np.sum(np.isnan(isfcs_loo_F)) ==
            np.sum(np.isnan(isfcs_loo_95)) == 1680)
    assert (np.sum(np.isnan(isfcs_loo_80)) ==
            np.sum(np.isnan(isfcs_loo_90)) == 1194)
    assert (np.sum(np.isnan(isfcs_loo_T)) ==
            np.sum(np.isnan(isfcs_loo_60)) ==
            np.sum(np.isnan(isfcs_loo_70)) == 802)
    assert np.array_equal(np.sum(np.isnan(isfcs_loo_F), axis=0),
                          np.sum(np.isnan(isfcs_loo_95), axis=0))
    assert np.array_equal(np.sum(np.isnan(isfcs_loo_80), axis=0),
                          np.sum(np.isnan(isfcs_loo_90), axis=0))
    assert np.all((np.array_equal(
                        np.sum(np.isnan(isfcs_loo_T), axis=0),
                        np.sum(np.isnan(isfcs_loo_60), axis=0)),
                   np.array_equal(
                        np.sum(np.isnan(isfcs_loo_T), axis=0),
                        np.sum(np.isnan(isfcs_loo_70), axis=0)),
                   np.array_equal(
                        np.sum(np.isnan(isfcs_loo_60), axis=0),
                        np.sum(np.isnan(isfcs_loo_70), axis=0))))

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    # Make sure voxel with NaNs across all subjects is always removed
    data[0, 0, :] = np.nan
    isfcs_loo_T = isfc(data, pairwise=False, vectorize_isfcs=False,
                       tolerate_nans=True)
    isfcs_loo_F = isfc(data, pairwise=False, vectorize_isfcs=False,
                       tolerate_nans=False)
    assert np.allclose(isfcs_loo_T, isfcs_loo_F, equal_nan=True)
    assert (np.sum(np.isnan(isfcs_loo_T)) ==
            np.sum(np.isnan(isfcs_loo_F)) ==
            1180)

    isfcs_pw_T = isfc(data, pairwise=True, vectorize_isfcs=False,
                      tolerate_nans=True)
    isfcs_pw_F = isfc(data, pairwise=True, vectorize_isfcs=False,
                      tolerate_nans=False)
    assert np.allclose(isfcs_pw_T, isfcs_pw_F, equal_nan=True)

    assert (np.sum(np.isnan(isfcs_pw_T)) ==
            np.sum(np.isnan(isfcs_pw_T)) ==
            11210)

    # Check for NaN-handling in targets
    n_targets = 15
    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)
    targets_data = simulated_timeseries(n_subjects, n_TRs,
                                        n_voxels=n_targets,
                                        data_type='array')

    # Inject NaNs into targets_data
    targets_data[0, 0, 0] = np.nan

    # Don't tolerate NaNs, should lose zeroeth voxel
    isfcs_loo = isfc(data,  targets=targets_data, pairwise=False,
                     vectorize_isfcs=False, tolerate_nans=False)
    assert np.sum(np.isnan(isfcs_loo)) == (n_subjects - 1) * (n_targets * 2)

    # Single NaN in targets will get averaged out with tolerate
    isfcs_loo = isfc(data, targets=targets_data, pairwise=False,
                     vectorize_isfcs=False, tolerate_nans=True)
    assert np.sum(np.isnan(isfcs_loo)) == 0


def test_squareform_isfc():

    # Set parameters for toy time series data
    n_subjects = 20
    n_TRs = 60
    n_voxels = 30
    random_state = 42

    logger.info("Testing ISC options")

    data = simulated_timeseries(n_subjects, n_TRs,
                                n_voxels=n_voxels, data_type='array',
                                random_state=random_state)

    # Generate square redundant ISFCs
    isfcs_r = isfc(data, vectorize_isfcs=False)
    assert isfcs_r.shape == (n_subjects, n_voxels, n_voxels)

    # Squareform these into condensed ISFCs and ISCs
    isfcs_c, iscs_c = squareform_isfc(isfcs_r)
    assert isfcs_c.shape == (n_subjects, n_voxels * (n_voxels - 1) / 2)
    assert iscs_c.shape == (n_subjects, n_voxels)

    # Go back the other way and check it's the same
    isfcs_new = squareform_isfc(isfcs_c, iscs_c)
    assert np.array_equal(isfcs_r, isfcs_new)

    # Check against ISC function
    assert np.allclose(isc(data), iscs_c, rtol=1e-03)

    # Check for two subjects
    isfcs_r = isfc(data[..., :2], vectorize_isfcs=False)
    assert isfcs_r.shape == (n_voxels, n_voxels)
    isfcs_c, iscs_c = squareform_isfc(isfcs_r)
    assert isfcs_c.shape == (n_voxels * (n_voxels - 1) / 2,)
    assert iscs_c.shape == (n_voxels,)
    assert np.array_equal(isfcs_r, squareform_isfc(isfcs_c, iscs_c))


if __name__ == '__main__':
    test_isc_input()
    test_isc_options()
    test_isc_output()
    test_isc_nans()
    test_bootstrap_isc()
    test_permutation_isc()
    test_timeshift_isc()
    test_phaseshift_isc()
    test_isfc_options()
    test_isfc_nans()
    test_squareform_isfc()
    logger.info("Finished all ISC tests")



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
import numpy as np
import scipy.stats as st
from brainiak.hyperparamopt.hpo import gmm_1d_distribution, fmin


def test_simple_gmm():
    x = np.array([1., 1., 2., 3., 1.])
    d = gmm_1d_distribution(x, min_limit=0., max_limit=4.)
    assert d(1.1) > d(3.5), "GMM distribution not behaving correctly"
    assert d(2.0) > d(3.0), "GMM distribution not behaving correctly"
    assert d(-1.0) == 0, "GMM distribution out of bounds error"
    assert d(9.0) == 0, "GMM distribution out of bounds error"

    samples = d.get_samples(n=25)
    np.testing.assert_array_less(samples, 4.)
    np.testing.assert_array_less(0., samples)


def test_simple_gmm_weights():
    x = np.array([1., 1., 2., 3., 1., 3.])
    d = gmm_1d_distribution(x)

    x2 = np.array([1., 2., 3.])
    w = np.array([3., 1., 2.])
    d2 = gmm_1d_distribution(x2, weights=w)
    y2 = d2(np.array([1.1, 2.0]))

    assert d2(1.1) == y2[0],\
        "GMM distribution array & scalar results don't match"
    assert np.abs(d(1.1) - d2(1.1)) < 1e-5,\
        "GMM distribution weights not handled correctly"
    assert np.abs(d(2.0) - d2(2.0)) < 1e-5,\
        "GMM distribution weights not handled correctly"


def test_simple_hpo():

    def f(args):
        x = args['x']
        return x*x

    s = {'x': {'dist': st.uniform(loc=-10., scale=20), 'lo': -10., 'hi': 10.}}
    trials = []

    # Test fmin and ability to continue adding to trials
    best = fmin(loss_fn=f, space=s, max_evals=40, trials=trials)
    best = fmin(loss_fn=f, space=s, max_evals=10, trials=trials)

    assert len(trials) == 50, "HPO continuation trials not working"

    # Test verbose flag
    best = fmin(loss_fn=f, space=s, max_evals=10, trials=trials)

    yarray = np.array([tr['loss'] for tr in trials])
    np.testing.assert_array_less(yarray, 100.)

    xarray = np.array([tr['x'] for tr in trials])
    np.testing.assert_array_less(np.abs(xarray), 10.)

    assert best['loss'] < 100., "HPO out of range"
    assert np.abs(best['x']) < 10., "HPO out of range"

    # Test unknown distributions
    s2 = {'x': {'dist': 'normal', 'mu': 0., 'sigma': 1.}}
    trials2 = []
    with pytest.raises(ValueError) as excinfo:
        fmin(loss_fn=f, space=s2, max_evals=40, trials=trials2)
    assert "Unknown distribution type for variable" in str(excinfo.value)

    s3 = {'x': {'dist': st.norm(loc=0., scale=1.)}}
    trials3 = []
    fmin(loss_fn=f, space=s3, max_evals=40, trials=trials3)



#  Copyright 2017 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import Iterable, Sequence

import nibabel as nib
import numpy as np
import pytest

from brainiak import io


@pytest.fixture
def in_dir() -> Path:
    return Path(__file__).parent / "data"


@pytest.fixture
def expected_image_data_shape() -> Sequence[int]:
    return (64, 64, 26, 10)


@pytest.fixture
def mask_path(in_dir: Path) -> Path:
    return in_dir / "mask.nii.gz"


@pytest.fixture
def labels_path(in_dir: Path) -> Path:
    return in_dir / "epoch_labels.npy"


@pytest.fixture
def expected_condition_spec_shape() -> Sequence[int]:
    return (2, 2, 10)


@pytest.fixture
def expected_n_subjects() -> int:
    return 2


@pytest.fixture
def image_paths(in_dir: Path) -> Iterable[Path]:
    return (in_dir / "subject1_bet.nii.gz", in_dir / "subject2_bet.nii.gz")


def test_load_images_from_dir_data_shape(
        in_dir: Path,
        expected_image_data_shape: Sequence[int],
        expected_n_subjects: int
        ) -> None:
    for i, image in enumerate(io.load_images_from_dir(in_dir, "bet.nii.gz")):
        assert image.get_fdata().shape == (64, 64, 26, 10)
    assert i + 1 == expected_n_subjects


def test_load_images_data_shape(
        image_paths: Iterable[Path],
        expected_image_data_shape: Sequence[int],
        expected_n_subjects: int
        ) -> None:
    for i, image in enumerate(io.load_images(image_paths)):
        assert image.get_fdata().shape == (64, 64, 26, 10)
    assert i + 1 == expected_n_subjects


def test_load_boolean_mask(mask_path: Path) -> None:
    mask = io.load_boolean_mask(mask_path)
    assert mask.dtype == bool


def test_load_boolean_mask_predicate(mask_path: Path) -> None:
    mask = io.load_boolean_mask(mask_path, lambda x: np.logical_not(x))
    expected_mask = np.logical_not(io.load_boolean_mask(mask_path))
    assert np.array_equal(mask, expected_mask)


def test_load_labels(labels_path: Path,
                     expected_condition_spec_shape: Sequence[int],
                     expected_n_subjects: int) -> None:
    condition_specs = io.load_labels(labels_path)
    i = 0
    for condition_spec in condition_specs:
        assert condition_spec.shape == expected_condition_spec_shape
        i += 1
    assert i == expected_n_subjects


def test_save_as_nifti_file(tmpdir) -> None:
    out_file = str(tmpdir / "nifti.nii")
    shape = (4, 4, 4)
    io.save_as_nifti_file(np.ones(shape), np.eye(4), out_file)
    assert nib.load(out_file).get_fdata().shape == shape        # type: ignore



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from brainiak.fcma.classifier import Classifier
from brainiak.fcma.preprocessing import prepare_fcma_data
from brainiak import io
from sklearn import svm
#from sklearn.linear_model import LogisticRegression
import sys
import logging
import numpy as np
from scipy.spatial.distance import hamming
from sklearn import model_selection
#from sklearn.externals import joblib

format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
# if want to output log to a file instead of outputting log to the console,
# replace "stream=sys.stdout" with "filename='fcma.log'"
logging.basicConfig(level=logging.INFO, format=format, stream=sys.stdout)
logger = logging.getLogger(__name__)

def example_of_aggregating_sim_matrix(raw_data, labels, num_subjects, num_epochs_per_subj):
    # aggregate the kernel matrix to save memory
    svm_clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    clf = Classifier(svm_clf, num_processed_voxels=1000, epochs_per_subj=num_epochs_per_subj)
    rearranged_data = raw_data[num_epochs_per_subj:] + raw_data[0:num_epochs_per_subj]
    rearranged_labels = labels[num_epochs_per_subj:] + labels[0:num_epochs_per_subj]
    clf.fit(list(zip(rearranged_data, rearranged_data)), rearranged_labels,
            num_training_samples=num_epochs_per_subj*(num_subjects-1))
    predict = clf.predict()
    print(predict)
    print(clf.decision_function())
    test_labels = labels[0:num_epochs_per_subj]
    incorrect_predict = hamming(predict, np.asanyarray(test_labels)) * num_epochs_per_subj
    logger.info(
        'when aggregating the similarity matrix to save memory, '
        'the accuracy is %d / %d = %.2f' %
        (num_epochs_per_subj-incorrect_predict, num_epochs_per_subj,
         (num_epochs_per_subj-incorrect_predict) * 1.0 / num_epochs_per_subj)
    )
    # when the kernel matrix is computed in portion, the test data is already in
    print(clf.score(None, test_labels))

def example_of_cross_validation_with_detailed_info(raw_data, labels, num_subjects, num_epochs_per_subj):
    # no shrinking, set C=1
    svm_clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    #logit_clf = LogisticRegression()
    clf = Classifier(svm_clf, epochs_per_subj=num_epochs_per_subj)
    # doing leave-one-subject-out cross validation
    for i in range(num_subjects):
        leave_start = i * num_epochs_per_subj
        leave_end = (i+1) * num_epochs_per_subj
        training_data = raw_data[0:leave_start] + raw_data[leave_end:]
        test_data = raw_data[leave_start:leave_end]
        training_labels = labels[0:leave_start] + labels[leave_end:]
        test_labels = labels[leave_start:leave_end]
        clf.fit(list(zip(training_data, training_data)), training_labels)
        # joblib can be used for saving and loading models
        #joblib.dump(clf, 'model/logistic.pkl')
        #clf = joblib.load('model/svm.pkl')
        predict = clf.predict(list(zip(test_data, test_data)))
        print(predict)
        print(clf.decision_function(list(zip(test_data, test_data))))
        incorrect_predict = hamming(predict, np.asanyarray(test_labels)) * num_epochs_per_subj
        logger.info(
            'when leaving subject %d out for testing, the accuracy is %d / %d = %.2f' %
            (i, num_epochs_per_subj-incorrect_predict, num_epochs_per_subj,
             (num_epochs_per_subj-incorrect_predict) * 1.0 / num_epochs_per_subj)
        )
        print(clf.score(list(zip(test_data, test_data)), test_labels))

def example_of_cross_validation_using_model_selection(raw_data, labels, num_subjects, num_epochs_per_subj):
    # NOTE: this method does not work for sklearn.svm.SVC with precomputed kernel
    # when the kernel matrix is computed in portions; also, this method only works
    # for self-correlation, i.e. correlation between the same data matrix.

    # no shrinking, set C=1
    svm_clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    #logit_clf = LogisticRegression()
    clf = Classifier(svm_clf, epochs_per_subj=num_epochs_per_subj)
    # doing leave-one-subject-out cross validation
    # no shuffling in cv
    skf = model_selection.StratifiedKFold(n_splits=num_subjects,
                                          shuffle=False)
    scores = model_selection.cross_val_score(clf, list(zip(raw_data, raw_data)),
                                             y=labels,
                                             cv=skf)
    print(scores)
    logger.info(
        'the overall cross validation accuracy is %.2f' %
        np.mean(scores)
    )

def example_of_correlating_two_components(raw_data, raw_data2, labels, num_subjects, num_epochs_per_subj):
    # aggregate the kernel matrix to save memory
    svm_clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    clf = Classifier(svm_clf, epochs_per_subj=num_epochs_per_subj)
    num_training_samples=num_epochs_per_subj*(num_subjects-1)
    clf.fit(list(zip(raw_data[0:num_training_samples], raw_data2[0:num_training_samples])),
            labels[0:num_training_samples])
    X = list(zip(raw_data[num_training_samples:], raw_data2[num_training_samples:]))
    predict = clf.predict(X)
    print(predict)
    print(clf.decision_function(X))
    test_labels = labels[num_training_samples:]
    incorrect_predict = hamming(predict, np.asanyarray(test_labels)) * num_epochs_per_subj
    logger.info(
        'when aggregating the similarity matrix to save memory, '
        'the accuracy is %d / %d = %.2f' %
        (num_epochs_per_subj-incorrect_predict, num_epochs_per_subj,
         (num_epochs_per_subj-incorrect_predict) * 1.0 / num_epochs_per_subj)
    )
    # when the kernel matrix is computed in portion, the test data is already in
    print(clf.score(X, test_labels))

def example_of_correlating_two_components_aggregating_sim_matrix(raw_data, raw_data2, labels,
                                                                 num_subjects, num_epochs_per_subj):
    # aggregate the kernel matrix to save memory
    svm_clf = svm.SVC(kernel='precomputed', shrinking=False, C=1, gamma='auto')
    clf = Classifier(svm_clf, num_processed_voxels=1000, epochs_per_subj=num_epochs_per_subj)
    num_training_samples=num_epochs_per_subj*(num_subjects-1)
    clf.fit(list(zip(raw_data, raw_data2)), labels,
            num_training_samples=num_training_samples)
    predict = clf.predict()
    print(predict)
    print(clf.decision_function())
    test_labels = labels[num_training_samples:]
    incorrect_predict = hamming(predict, np.asanyarray(test_labels)) * num_epochs_per_subj
    logger.info(
        'when aggregating the similarity matrix to save memory, '
        'the accuracy is %d / %d = %.2f' %
        (num_epochs_per_subj-incorrect_predict, num_epochs_per_subj,
         (num_epochs_per_subj-incorrect_predict) * 1.0 / num_epochs_per_subj)
    )
    # when the kernel matrix is computed in portion, the test data is already in
    print(clf.score(None, test_labels))

# python3 classification.py face_scene bet.nii.gz face_scene/prefrontal_top_mask.nii.gz face_scene/fs_epoch_labels.npy
if __name__ == '__main__':
    if len(sys.argv) != 5:
        logger.error('the number of input argument is not correct')
        sys.exit(1)

    data_dir = sys.argv[1]
    extension = sys.argv[2]
    mask_file = sys.argv[3]
    epoch_file = sys.argv[4]

    epoch_list = np.load(epoch_file)
    num_subjects = len(epoch_list)
    num_epochs_per_subj = epoch_list[0].shape[1]

    images = io.load_images_from_dir(data_dir, extension)
    mask = io.load_boolean_mask(mask_file)
    conditions = io.load_labels(epoch_file)
    raw_data, _, labels = prepare_fcma_data(images, conditions, mask)

    example_of_aggregating_sim_matrix(raw_data, labels, num_subjects, num_epochs_per_subj)

    example_of_cross_validation_with_detailed_info(raw_data, labels, num_subjects, num_epochs_per_subj)

    example_of_cross_validation_using_model_selection(raw_data, labels, num_subjects, num_epochs_per_subj)

    # test of two different components for correlation computation
    # images = io.load_images_from_dir(data_dir, extension)
    # mask2 = io.load_boolean_mask('face_scene/visual_top_mask.nii.gz')
    # raw_data, raw_data2, labels = prepare_fcma_data(images, conditions, mask,
    #                                                 mask2)
    #example_of_correlating_two_components(raw_data, raw_data2, labels, num_subjects, num_epochs_per_subj)
    #example_of_correlating_two_components_aggregating_sim_matrix(raw_data, raw_data2, labels, num_subjects, num_epochs_per_subj)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import namedtuple

import numpy as np
from mpi4py import MPI

import pytest

from brainiak.searchlight.searchlight import Searchlight
from brainiak.searchlight.searchlight import Diamond, Ball


"""Distributed Searchlight Test
"""


def cube_sfn(data, msk, myrad, bcast_var):
    if np.all(msk) and np.any(msk):
        return 1.0
    return None


@pytest.mark.mpiexec(n=2)
def test_searchlight_with_cube(pool_size):
    sl = Searchlight(sl_rad=3)
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size
    dim0, dim1, dim2 = (50, 50, 50)
    ntr = 30
    nsubj = 3
    mask = np.zeros((dim0, dim1, dim2), dtype=bool)
    data = [np.empty((dim0, dim1, dim2, ntr), dtype=object)
            if i % size == rank
            else None
            for i in range(0, nsubj)]

    # Put a spot in the mask
    mask[10:17, 10:17, 10:17] = True

    sl.distribute(data, mask)
    global_outputs = sl.run_searchlight(cube_sfn, pool_size=pool_size)

    if rank == 0:
        assert global_outputs[13, 13, 13] == 1.0
        global_outputs[13, 13, 13] = None

        for i in range(global_outputs.shape[0]):
            for j in range(global_outputs.shape[1]):
                for k in range(global_outputs.shape[2]):
                    assert global_outputs[i, j, k] is None


@pytest.mark.mpiexec(n=2)
def test_searchlight_with_cube_poolsize_1():
    sl = Searchlight(sl_rad=3)
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size
    dim0, dim1, dim2 = (50, 50, 50)
    ntr = 30
    nsubj = 3
    mask = np.zeros((dim0, dim1, dim2), dtype=bool)
    data = [np.empty((dim0, dim1, dim2, ntr), dtype=object)
            if i % size == rank
            else None
            for i in range(0, nsubj)]

    # Put a spot in the mask
    mask[10:17, 10:17, 10:17] = True

    sl.distribute(data, mask)
    global_outputs = sl.run_searchlight(cube_sfn, pool_size=1)

    if rank == 0:
        assert global_outputs[13, 13, 13] == 1.0
        global_outputs[13, 13, 13] = None

        for i in range(global_outputs.shape[0]):
            for j in range(global_outputs.shape[1]):
                for k in range(global_outputs.shape[2]):
                    assert global_outputs[i, j, k] is None


def diamond_sfn(data, msk, myrad, bcast_var):
    assert not np.any(msk[~Diamond(3).mask_])
    if np.all(msk[Diamond(3).mask_]):
        return 1.0
    return None


@pytest.mark.mpiexec(n=2)
def test_searchlight_with_diamond(pool_size):
    sl = Searchlight(sl_rad=3, shape=Diamond)
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size
    dim0, dim1, dim2 = (50, 50, 50)
    ntr = 30
    nsubj = 3
    mask = np.zeros((dim0, dim1, dim2), dtype=bool)
    data = [np.empty((dim0, dim1, dim2, ntr), dtype=object)
            if i % size == rank
            else None
            for i in range(0, nsubj)]

    # Put a spot in the mask
    mask[10:17, 10:17, 10:17] = Diamond(3).mask_

    sl.distribute(data, mask)
    global_outputs = sl.run_searchlight(diamond_sfn, pool_size=pool_size)

    if rank == 0:
        assert global_outputs[13, 13, 13] == 1.0
        global_outputs[13, 13, 13] = None

        for i in range(global_outputs.shape[0]):
            for j in range(global_outputs.shape[1]):
                for k in range(global_outputs.shape[2]):
                    assert global_outputs[i, j, k] is None


def ball_sfn(data, msk, myrad, bcast_var):
    x, y, z = np.mgrid[-myrad:myrad+1, -myrad:myrad+1, -myrad:myrad+1]
    correct_mask = np.square(x) + np.square(y) + np.square(z) <= myrad ** 2
    assert not np.any(msk[~Ball(3).mask_])
    if np.all(correct_mask == msk):
        return 1.0
    return None


@pytest.mark.mpiexec(n=2)
def test_searchlight_with_ball(pool_size):
    sl = Searchlight(sl_rad=3, shape=Ball)
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size
    dim0, dim1, dim2 = (50, 50, 50)
    ntr = 30
    nsubj = 3
    mask = np.zeros((dim0, dim1, dim2), dtype=bool)
    data = [np.empty((dim0, dim1, dim2, ntr), dtype=object)
            if i % size == rank
            else None
            for i in range(0, nsubj)]

    # Put a spot in the mask
    mask[10:17, 10:17, 10:17] = Ball(3).mask_

    sl.distribute(data, mask)
    global_outputs = sl.run_searchlight(ball_sfn, pool_size=pool_size)

    if rank == 0:
        assert global_outputs[13, 13, 13] == 1.0
        global_outputs[13, 13, 13] = None

        for i in range(global_outputs.shape[0]):
            for j in range(global_outputs.shape[1]):
                for k in range(global_outputs.shape[2]):
                    assert global_outputs[i, j, k] is None


MaskRadBcast = namedtuple("MaskRadBcast", "mask rad")


def test_instantiate():
    sl = Searchlight(sl_rad=5, max_blk_edge=10)
    assert sl


def voxel_test_sfn(data, msk, myrad, bcast):
    rad = bcast.rad
    # Check each point
    for subj in data:
        for _tr in range(subj.shape[3]):
            tr = subj[:, :, :, _tr]
            midpt = tr[rad, rad, rad]
            for d0 in range(tr.shape[0]):
                for d1 in range(tr.shape[1]):
                    for d2 in range(tr.shape[2]):
                        assert np.array_equal(tr[d0, d1, d2] - midpt,
                                              np.array([d0-rad, d1-rad,
                                                        d2-rad, 0]))

    # Determine midpoint
    midpt = data[0][rad, rad, rad, 0]
    midpt = (midpt[0], midpt[1], midpt[2])

    for d0 in range(msk.shape[0]):
        for d1 in range(msk.shape[1]):
            for d2 in range(msk.shape[2]):
                pt = (midpt[0] - rad + d0, midpt[1] - rad + d1,
                      midpt[2] - rad + d2)
                assert bcast.mask[pt] == msk[d0, d1, d2]

    # Return midpoint
    return midpt


def block_test_sfn(data, msk, myrad, bcast_var, extra_params):
    outmat = data[0][:, :, :, 0]
    outmat[~msk] = None
    if myrad == 0:
        return outmat
    else:
        return outmat[myrad:-myrad, myrad:-myrad, myrad:-myrad]


@pytest.mark.mpiexec(n=2, timeout=120)
def test_correctness(pool_size):  # noqa: C901
    def voxel_test(data, mask, max_blk_edge, rad):

        comm = MPI.COMM_WORLD
        rank = comm.rank

        (dim0, dim1, dim2) = mask.shape

        # Initialize dataset with known pattern
        for subj in data:
            if subj is not None:
                for tr in range(subj.shape[3]):
                    for d1 in range(dim0):
                        for d2 in range(dim1):
                            for d3 in range(dim2):
                                subj[d1, d2, d3, tr] = np.array(
                                    [d1, d2, d3, tr])

        sl = Searchlight(sl_rad=rad, max_blk_edge=max_blk_edge)
        sl.distribute(data, mask)
        sl.broadcast(MaskRadBcast(mask, rad))
        global_outputs = sl.run_searchlight(voxel_test_sfn,
                                            pool_size=pool_size)

        if rank == 0:
            for d0 in range(rad, global_outputs.shape[0]-rad):
                for d1 in range(rad, global_outputs.shape[1]-rad):
                    for d2 in range(rad, global_outputs.shape[2]-rad):
                        if mask[d0, d1, d2]:
                            assert np.array_equal(
                                np.array(global_outputs[d0, d1, d2]),
                                np.array([d0, d1, d2]))

    def block_test(data, mask, max_blk_edge, rad):

        comm = MPI.COMM_WORLD
        rank = comm.rank

        (dim0, dim1, dim2) = mask.shape

        # Initialize dataset with known pattern
        for subj in data:
            if subj is not None:
                for tr in range(subj.shape[3]):
                    for d1 in range(dim0):
                        for d2 in range(dim1):
                            for d3 in range(dim2):
                                subj[d1, d2, d3, tr] = np.array(
                                    [d1, d2, d3, tr])

        sl = Searchlight(sl_rad=rad, max_blk_edge=max_blk_edge)
        sl.distribute(data, mask)
        sl.broadcast(mask)
        global_outputs = sl.run_block_function(block_test_sfn,
                                               pool_size=pool_size)

        if rank == 0:
            for d0 in range(rad, global_outputs.shape[0]-rad):
                for d1 in range(rad, global_outputs.shape[1]-rad):
                    for d2 in range(rad, global_outputs.shape[2]-rad):
                        if mask[d0, d1, d2]:
                            assert np.array_equal(
                                np.array(global_outputs[d0, d1, d2]),
                                np.array([d0, d1, d2, 0]))

    # Create dataset
    def do_test(dim0, dim1, dim2, ntr, nsubj, max_blk_edge, rad):
        comm = MPI.COMM_WORLD
        rank = comm.rank
        size = comm.size
        mask = np.random.choice([True, False], (dim0, dim1, dim2))
        data = [np.empty((dim0, dim1, dim2, ntr), dtype=object)
                if i % size == rank
                else None
                for i in range(0, nsubj)]
        voxel_test(data, mask, max_blk_edge, rad)
        block_test(data, mask, max_blk_edge, rad)

    do_test(dim0=7, dim1=5, dim2=9, ntr=5, nsubj=1, max_blk_edge=4, rad=1)
    do_test(dim0=7, dim1=5, dim2=9, ntr=5, nsubj=1, max_blk_edge=4, rad=0)
    do_test(dim0=7, dim1=5, dim2=9, ntr=5, nsubj=5, max_blk_edge=4, rad=1)
    do_test(dim0=1, dim1=5, dim2=9, ntr=5, nsubj=5, max_blk_edge=4, rad=1)
    do_test(dim0=0, dim1=10, dim2=8, ntr=5, nsubj=5, max_blk_edge=4, rad=1)
    do_test(dim0=7, dim1=5, dim2=9, ntr=5, nsubj=1, max_blk_edge=4, rad=2)
    do_test(dim0=7, dim1=5, dim2=9, ntr=5, nsubj=1, max_blk_edge=4, rad=3)



#  Copyright 2018 David Huberdeau & Peter Kok
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Copyright 2020 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Inverted Encoding Model (IEM)

    Method to decode and reconstruct features from data.

    The implementation is roughly based on the following publications:

    [Kok2013] "1.Kok, P., Brouwer, G. J., Gerven, M. A. J. van &
    Lange, F. P. de. Prior Expectations Bias Sensory Representations
    in Visual Cortex. J. Neurosci. 33, 16275–16284 (2013).

    [Brouwer2011] "2.Brouwer, G. J. & Heeger, D. J. Cross-orientation
    suppression in human visual cortex. J. Neurophysiol. 106(5):
    2108-2119 (2011).

    [Brouwer2009] "3.Brouwer, G. J. & Heeger, D. J.
    Decoding and Reconstructing Color from Responses in Human Visual
    Cortex. J. Neurosci. 29, 13992–14003 (2009).

    This implementation uses a set of sinusoidal
    basis functions to represent the set of possible feature values.
    A feature value is some characteristic of a stimulus, e.g. the
    angular location of a target along a horizontal line. This code was
    written to give some flexibility compared to the specific instances
    in Kok, 2013 & in Brouwer, 2009. Users can set the number of basis
    functions, or channels, and the range of possible feature values.

    There are separate classes for reconstructing feature values in a
    1-dimensional (1D) space or in a 2-dimensional (2D) space.
"""

# Authors: David Huberdeau (Yale University) &
# Peter Kok (Yale University), 2018 &
# Vy Vo (Intel Corp., UCSD), 2019

import logging
import warnings
import numpy as np
import scipy.stats
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import cosine_distances, euclidean_distances
from ..utils.utils import circ_dist

__all__ = ["InvertedEncoding1D",
           "InvertedEncoding2D"]

logger = logging.getLogger(__name__)
MAX_CONDITION_CHECK = 9000


class InvertedEncoding1D(BaseEstimator):
    """Basis function-based reconstruction method

    Inverted encoding models (alternatively known as forward models) are used
    to reconstruct a feature represented in some N-dimensional space, here 1D,
    (e.g. color of a stimulus) from patterns across voxels in functional data.
    The model uses n_channels number of idealized basis functions and assumes
    that the transformation from stimulus feature (e.g. color) to basis
    function is one- to-one and invertible. The response of a voxel is
    expressed as the weighted sum of basis functions. In this implementation,
    basis functions were half-wave rectified sinusoid functions raised to a
    power set by the user (e.g. 6).

    The model:
    Inverted encoding models reconstruct a stimulus feature from
    patterns of BOLD activity by relating the activity in each
    voxel, B, to the values of hypothetical channels (or basis
    functions), C, according to Equation 1 below.

    (1)     B = W*C

    where W is a weight matrix that represents the relationship
    between BOLD activity and Channels. W must be estimated from
    training data; this implementation (and most described in the
    literature) uses linear regression to estimate W as in Equation
    2 below [note: inv() represents matrix inverse or
    pseudo-inverse].

    (2)     W_est = B_train*inv(C_train)

    The weights in W_est (short for "estimated") represent the
    contributions of each channel to the response of each voxel.
    Estimated channel responses can be computed given W_est and
    new voxel activity represented in matrix B_exp (short for
    "experiment") through inversion of Equation 1:

    (3)     C_est = inv(W_est)*B_exp

    Given estimated channel responses, C_est, it is straightforward
    to obtain the reconstructed feature value by summing over
    channels multiplied by their channel responses and taking the
    argmax (i.e. the feature associated with the maximum value).

    Using this model:
    Use fit() to estimate the weights of the basis functions given
    input data (e.g. beta values from fMRI data). This function
    will execute equation 2 above.

    Use predict() to compute predicted stimulus values
    from new functional data. This function computes estimated
    channel responses, as in equation 3, then computes summed
    channel output and finds the argmax (within the stimulus
    feature space) associated with those responses.

    Use score() to compute a measure of the error of the prediction
    based on known stimuli.

    This implementation assumes a circular (or half-
    circular) feature domain. Future implementations might
    generalize the feature input space, and increase the
    possible dimensionality.

    Parameters
    ----------
    n_channels: int, default 5. Number of channels
        The number of channels, or basis functions, to be used in
        the inverted encoding model.

    channel_exp: int, default 6. Basis function exponent.
        The exponent of the sinuoidal basis functions, which
        establishes the width of the functions.

    stimulus_mode: str, default 'halfcircular' (other option is
        'circular'). Describes the feature domain.

    range_start: double, default 0. Lowest value of domain.
        Beginning value of range of independent variable
        (usually degrees).

    range_stop: double, default 180. Highest value of domain.
        Ending value of range of independent variable
        (usually degrees).

    channel_density: int, default 180. Number of points in the
        feature domain.

    stimulus_resolution: double, default None will set the stimulus
        resolution to be identical to the channel density. This sets
        the resolution at which the stimuli were presented (e.g. a
        spatial position with some width has a lower stimulus
        resolution).

    Attributes
    ----------
    channels_: [n_channels, channel density] NumPy 2D array
        matrix defining channel values

    W_: sklearn.linear_model model containing weight matrix that
        relates estimated channel responses to response amplitude
        data

    See get_params() for the rest of the attributes.
    """

    def __init__(self, n_channels=6, channel_exp=5,
                 stimulus_mode='halfcircular', range_start=0., range_stop=180.,
                 channel_density=180, stimulus_resolution=None):
        self.n_channels = n_channels
        self.channel_exp = channel_exp
        self.stimulus_mode = stimulus_mode
        self.range_start = range_start
        self.range_stop = range_stop
        self.channel_density = channel_density
        self.channel_domain = np.linspace(range_start, range_stop - 1,
                                          channel_density)
        if stimulus_resolution is None:
            self.stim_res = channel_density
        else:
            self.stim_res = stimulus_resolution
        self._check_params()

    def _check_params(self):
        if self.range_start >= self.range_stop:
            raise ValueError("range_start {} must be less than "
                             "{} range_stop.".format(self.range_start,
                                                     self.range_stop))
        if self.stimulus_mode == 'halfcircular':
            if (self.range_stop - self.range_start) != 180.:
                raise ValueError("For half-circular feature spaces,"
                                 "the range must be 180 degrees, "
                                 "not {}".
                                 format(self.range_stop - self.range_start))
        elif self.stimulus_mode == 'circular':
            if (self.range_stop - self.range_start) != 360.:
                raise ValueError("For circular feature spaces, the"
                                 " range must be 360 degrees"
                                 "not {}".
                                 format(self.range_stop - self.range_start))
        if self.n_channels < 2:
            raise ValueError("Insufficient number of channels.")
        if not np.isin(self.stimulus_mode, ['circular', 'halfcircular']):
            raise ValueError("Stimulus mode must be one of these: "
                             "'circular', 'halfcircular'")

    def fit(self, X, y):
        """Use data and feature variable labels to fit an IEM

        Parameters
        ----------
        X: numpy matrix of voxel activation data. [observations, voxels]
            Should contain the beta values for each observation or
            trial and each voxel of training data.
        y: numpy array of response variable. [observations]
            Should contain the feature for each observation in X.
        """
        # Check that data matrix is well conditioned:
        if np.linalg.cond(X) > MAX_CONDITION_CHECK:
            logger.error("Data is singular.")
            raise ValueError("Data matrix is nearly singular.")
        if X.shape[0] < self.n_channels:
            logger.error("Not enough observations. Cannot calculate "
                         "pseudoinverse.")
            raise ValueError("Fewer observations (trials) than "
                             "channels. Cannot compute pseudoinverse.")
        # Check that the data matrix is the right size
        shape_data = np.shape(X)
        shape_labels = np.shape(y)
        if len(shape_data) != 2:
            raise ValueError("Data matrix has too many or too few "
                             "dimensions.")
        else:
            if shape_data[0] != shape_labels[0]:
                raise ValueError("Mismatched data samples and label samples")

        # Define the channels (or basis set)
        self.channels_, channel_centers = self._define_channels()
        logger.info("Defined channels centered at {} degrees.".format(
            np.rad2deg(channel_centers)))
        # Create a matrix of channel activations for every observation.
        # (i.e., C1 in Brouwer & Heeger 2009.)
        C = self._define_trial_activations(y)
        # Solve for W in B = WC
        self.W_ = X.transpose() @ np.linalg.pinv(C.transpose())
        if np.linalg.cond(self.W_) > MAX_CONDITION_CHECK:
            logger.error("Weight matrix is nearly singular.")
            raise ValueError("Weight matrix is nearly singular.")

        return self

    def predict(self, X):
        """Use test data to predict the feature

        Parameters
        ----------
            X: numpy matrix of voxel activation from test trials
            [observations, voxels]. Used to predict feature
            associated with the given observation.

        Returns
        -------
            model_prediction: numpy array of estimated feature values.
        """
        # Check that the data matrix is the right size
        shape_data = np.shape(X)
        if len(shape_data) != 2:
            raise ValueError("Data matrix has too many or too few "
                             "dimensions.")

        model_prediction = self._predict_features(X)

        return model_prediction

    def score(self, X, y):
        """Calculate error measure of prediction. Default measurement
        is R^2, the coefficient of determination.

        Parameters
        ----------
            X: numpy matrix of voxel activation from new data
                [observations,voxels]
            y: numpy array of responses. [observations]

        Returns
        -------
            score_value: the error measurement between the actual
                feature and predicted features.
        """
        pred_features = self.predict(X)
        if self.stimulus_mode == 'halfcircular':
            # multiply features by 2. otherwise doesn't wrap properly
            pred_features = pred_features * 2
            y = y * 2

        ssres = (circ_dist(np.deg2rad(y),
                           np.deg2rad(pred_features)) ** 2).sum()
        sstot = (circ_dist(np.deg2rad(y),
                           np.ones(y.size) * scipy.stats.circmean(
                               np.deg2rad(y))) ** 2).sum()
        score_value = (1 - ssres / sstot)

        return score_value

    def get_params(self, deep: bool = True):
        """Returns model parameters.

        Returns
        -------
        params: parameter of this object
        """
        return {"n_channels": self.n_channels, "channel_exp": self.channel_exp,
                "stimulus_mode": self.stimulus_mode,
                "range_start": self.range_start, "range_stop": self.range_stop,
                "channel_domain": self.channel_domain,
                "stim_res": self.stim_res}

    def set_params(self, **parameters):
        """Sets model parameters after initialization.

        Parameters
        ----------
            parameters: structure with parameters and change values
        """
        for parameter, value in parameters.items():
            setattr(self, parameter, value)

        setattr(self, "channel_domain",
                np.linspace(self.range_start, self.range_stop - 1,
                            self.channel_density))
        self._check_params()
        return self

    def _define_channels(self):
        """Define basis functions (aka channels).

        Returns
        -------
            channels: numpy matrix of basis functions. dimensions are
                [n_channels, function resolution].
            channel_centers: numpy array of the centers of each channel
        """
        channel_centers = np.linspace(np.deg2rad(self.range_start),
                                      np.deg2rad(self.range_stop),
                                      self.n_channels + 1)
        channel_centers = channel_centers[0:-1]
        # make sure channels are not bimodal if using 360 deg space
        if self.stimulus_mode == 'circular':
            domain = self.channel_domain * 0.5
            centers = channel_centers * 0.5
        elif self.stimulus_mode == 'halfcircular':
            domain = self.channel_domain
            centers = channel_centers

        # define exponentiated function
        channels = np.asarray(
            [np.cos(np.deg2rad(domain) - cx) ** self.channel_exp for cx in
             centers])
        # half-wave rectification preserving circularity
        channels = abs(channels)

        return channels, channel_centers

    def _define_trial_activations(self, stimuli):
        """Defines a numpy matrix of predicted channel responses for
        each trial/observation.

        Parameters

            stimuli: numpy array of the feature values for each
                observation

        Returns
        -------
            C: matrix of predicted channel responses. dimensions are
                number of observations by stimulus resolution
        """
        stim_axis = np.linspace(self.range_start, self.range_stop - 1,
                                self.stim_res)
        if self.range_start > 0:
            stimuli = stimuli + self.range_start
        elif self.range_start < 0:
            stimuli = stimuli - self.range_start
        one_hot = np.eye(self.stim_res)
        indices = [np.argmin(abs(stim_axis - x)) for x in stimuli]
        stimulus_mask = one_hot[indices, :]
        if self.channel_density != self.stim_res:
            if self.channel_density % self.stim_res == 0:
                stimulus_mask = np.repeat(stimulus_mask,
                                          self.channel_density / self.stim_res)
            else:
                raise NotImplementedError("This code doesn't currently support"
                                          " stimuli which are not square "
                                          "functions in the feature domain, or"
                                          " stimulus widths that are not even"
                                          "divisors of the number of points in"
                                          " the feature domain.")

        C = stimulus_mask @ self.channels_.transpose()
        # Check that C is full rank
        if np.linalg.matrix_rank(C) < self.n_channels:
            warnings.warn("Stimulus matrix is {}, not full rank. May cause "
                          "issues with stimulus prediction/reconstruction.".
                          format(np.linalg.matrix_rank(C)), RuntimeWarning)
        return C

    def _predict_channel_responses(self, X):
        """Computes predicted channel responses from data
        (e.g. C2 in Brouwer & Heeger 2009)

        Parameters
        ----------
            X: numpy data matrix. [observations, voxels]

        Returns
        -------
            channel_response: numpy matrix of channel responses
        """
        channel_response = np.matmul(np.linalg.pinv(self.W_), X.transpose())
        return channel_response

    def _predict_feature_responses(self, X):
        """Takes channel weights and transforms them into continuous
        functions defined in the feature domain.

        Parameters
         ---------
            X: numpy matrix of data. [observations, voxels]

        Returns
        -------
            pred_response: predict response from all channels. Used
                to predict feature (e.g. direction).
        """
        pred_response = np.matmul(self.channels_.transpose(),
                                  self._predict_channel_responses(X))
        return pred_response

    def _predict_features(self, X):
        """Predicts feature value (e.g. direction) from data in X.
        Takes the maximum of the 'reconstructed' or predicted response
        function.

        Parameters
         ---------
            X: numpy matrix of data. [observations, voxels]

        Returns
        -------
            pred_features: predicted feature from response across all
                channels.
        """
        pred_response = self._predict_feature_responses(X)
        feature_ind = np.argmax(pred_response, 0)
        pred_features = self.channel_domain[feature_ind]

        return pred_features


class InvertedEncoding2D(BaseEstimator):
    """Basis function-based reconstruction method

    Inverted encoding models (alternatively known as forward models) are used
    to reconstruct a feature represented in a N-dimensional space, here 2D,
    (e.g. position on a projector screen) from patterns across voxels in
    functional data. The model uses some number of idealized basis functions
    that cover the 2D space, and assumes that the transformation from
    stimulus feature (e.g. 2D spatial position) to basis function is one-
    to-one and invertible. The response of a voxel is expressed as the
    weighted sum of basis functions. In this implementation, basis functions
    were half-wave rectified sinusoid functions raised to some power (set by
    the user).

    The documentation will refer to the 'stimulus space' or 'stimulus domain',
    which should be a 2D space in consistent units (e.g. screen pixels,
    or degrees visual angle). The stimulus space is the domain in which the
    stimulus is reconstructed. We will refer to the each point in this 2D
    stimulus domain as a 'pixel'.

    The model:
    Inverted encoding models reconstruct a stimulus feature from
    patterns of BOLD activity by relating the activity in each
    voxel, B, to the values of hypothetical channels (or basis
    functions), C, according to Equation 1 below.

    (1)     B = W*C

    where W is a weight matrix that represents the relationship
    between BOLD activity and Channels. W must be estimated from
    training data; this implementation (and most described in the
    literature) uses linear regression to estimate W as in Equation
    2 below [note: inv() represents matrix inverse or
    pseudo-inverse].

    (2)     W_est = B_train*inv(C_train)

    The weights in W_est (short for "estimated") represent the
    contributions of each channel to the response of each voxel.
    Estimated channel responses can be computed given W_est and
    new voxel activity represented in matrix B_exp (short for
    "experiment") through inversion of Equation 1:

    (3)     C_est = inv(W_est)*B_exp

    Given estimated channel responses, C_est, it is straightforward
    to obtain the reconstructed feature value by summing over
    channels multiplied by their channel responses and taking the
    argmax (i.e. the feature associated with the maximum value).

    Using this model:
    Use fit() to estimate the weights of the basis functions given
    input data (e.g. beta values from fMRI data). This function
    will execute equation 2 above.

    Use predict() to compute predicted stimulus values
    from new functional data. This function computes estimated
    channel responses, as in equation 3, then computes summed
    channel output and finds the argmax (within the stimulus
    feature space) associated with those responses.

    Use score() to compute a measure of the error of the prediction
    based on known stimuli.

    Parameters
    ----------
    stim_xlim: list of 2 floats Specifies the minimum and maximum x-values
        of the area to be reconstructed. In order to be estimated properly, a
        stimulus must appear at these limits. Specifying limits outside the
        range of the stimuli can lead to spurious estimates.

    stim_ylim: list of 2 floats Specifies the minimum and maximum y-values
        of the area to be reconstructed. In order to be estimated properly, a
        stimulus must appear at these limits. Specifying limits outside the
        range of the stimuli can lead to spurious estimates.

    stimulus_resolution: float or list of 2 floats. If a single float is
        given, it will be expanded to a list (i.e. we will assume that the
        reconstructed area is composed of square pixels).

    stim_radius: float, or sequence of floats [n_stim], default None. If the
        user does not define the design matrix of the encoding model (e.g. C
        in B = W*C), it will be defined automatically on the assumption that
        each observation was for a 2D circular stimulus of some radius.

    chan_xlim: list of 2 floats, default None. Specifies the minimum and
        maximum x-values of the channels, or basis functions.

    chan_ylim: list of 2 floats, default None. Specifies the minimum and
        maximum y-values of the channels, or basis functions.

    channels: [n_channels, n_pixels] NumPy 2D array, default None. If None at
        initialization, it can be defined with
        either define_basis_functions_sqgrid() or
        define_basis_functions_trigrid(), each of which tiles the given 2D
        space with some grid (square or triangular/hexagonal, respectively).
        Alternatively, the user can specify their own channels.

    channel_exp: int, default 7. Basis function exponent. The exponent of the
        sinuoidal basis functions, which helps control their width.

    Attributes
    ----------
    channels: [n_channels, n_pixels] NumPy 2D array defining channels

    W_: sklearn.linear_model containing weight matrix that relates estimated
        channel responses to response data

    See get_params() for the rest of the attributes.
    """

    def __init__(self, stim_xlim, stim_ylim, stimulus_resolution,
                 stim_radius=None, chan_xlim=None, chan_ylim=None,
                 channels=None, channel_exp=7):
        """Defines a 2D inverted encoding model object.

        While the parameters defining the domain in which to reconstruct
        the stimuli are required (e.g. all `stim_*` inputs), the parameters
        to define the channels (`chan*`) are optional, in case the user
        wishes to define their own channels (a.k.a basis functions).


        Parameters
        ----------
        stim_xlim: sequence of 2 float values, specifying the lower & upper
            limits on the horizontal axis, respectively.
        stim_ylim: sequence of 2 float values, specifying the lower & upper
            limits on the vertical axis, respectively.
        stimulus_resolution: a float or sequence of 2 floats, specifying the
            number of pixels that exist in the x- and y- directions.
        stim_radius: float, default None. The radius in pixels, assuming that
            the stimulus is circular. If None, the user must either define it
            before running fit(), or pass in a custom C in B = W*C.
        chan_xlim: sequence of 2 float values, default None. Specifies the
            lower & upper limits of the channels in the horizontal axis. If
            None, the user must define this before using the class functions
            to create basis functions, or pass in custom-defined channels.
        chan_ylim: sequence of 2 float values, default None. Specifies the
            lower & upper limits of the channels in the vertical axis. If
            None, the user must define this before using the class functions
            to create basis functions, or pass in custom-defined channels.
        channel_exp: float or int, default None. The exponent for a
            sinusoidal basis function. If None, it must be set before the
            channels or defined, or pass in custom-defined channels.

        """
        # Automatically expand stimulus_resolution if only one value is given.
        # This will create a square field  of view (FOV) for the
        # reconstruction.
        if not isinstance(stimulus_resolution, list):  # make FOV square
            stimulus_resolution = [stimulus_resolution, stimulus_resolution]
        if (len(stim_xlim) != 2) or (len(stim_ylim) != 2):
            raise ValueError("Stimulus limits should be a sequence, 2 values")
        self.stim_fov = [stim_xlim, stim_ylim]
        self.stim_pixels = [np.linspace(stim_xlim[0], stim_xlim[1],
                                        stimulus_resolution[0]),
                            np.linspace(stim_ylim[0], stim_ylim[1],
                                        stimulus_resolution[1])]
        self.xp, self.yp = np.meshgrid(self.stim_pixels[0],
                                       self.stim_pixels[1])
        self.stim_radius_px = stim_radius
        self.channels = channels
        if self.channels is None:
            self.n_channels = None
        else:
            self.n_channels = self.channels.shape[0]
        if chan_xlim is None:
            chan_xlim = stim_xlim
            logger.info("Set channel x-limits to stimulus x-limits", stim_xlim)
        if chan_ylim is None:
            chan_ylim = stim_ylim
            logger.info("Set channel y-limits to stimulus y-limits", stim_ylim)
        self.channel_limits = [chan_xlim, chan_ylim]
        self.channel_exp = channel_exp
        self._check_params()

    def _check_params(self):
        if len(self.stim_fov) != 2:
            raise ValueError("Stim FOV needs to have an x-list and a y-list")
        elif len(self.stim_fov[0]) != 2 or len(self.stim_fov[1]) != 2:
            raise ValueError("Stimulus limits should be a sequence, 2 values")
        else:
            if (self.stim_fov[0][0] >= self.stim_fov[0][1]) or \
                    (self.stim_fov[1][0] >= self.stim_fov[1][1]):
                raise ValueError("Stimulus x or y limits should be ascending "
                                 "values")
        if self.xp.size != self.yp.size:
            raise ValueError("xpixel grid and ypixel grid do not have same "
                             "number of elements")
        if self.n_channels and np.all(self.channels):
            if self.n_channels != self.channels.shape[0]:
                raise ValueError("Number of channels {} does not match the "
                                 "defined channels: {}".
                                 format(self.n_channels,
                                        self.channels.shape[0]))
            if self.channels.shape[1] != self.xp.size:
                raise ValueError("Defined {} channels over {} pixels, but "
                                 "stimuli are represented over {} pixels. "
                                 "Pixels should match.".
                                 format(self.n_channels,
                                        self.channels.shape[1],
                                        self.xp.size))

    def fit(self, X, y, C=None):
        """Use data and feature variable labels to fit an IEM

        Parameters
        ----------
        X: numpy matrix of voxel activation data. [observations, voxels]
            Should contain the beta values for each observation or
            trial and each voxel of training data.
        y: numpy array of response variable. [observations]
            Should contain the feature for each observation in X.
        C: numpy matrix of channel activations for every observation (e.g.
            the design matrix C in the linear equation B = W*C), matrix size
            [observations, pixels]. If None (default), this assumes that each
            observation contains a 2D circular stimulus and will define the
            activations with self._define_trial_activations(y).
        """
        # Check that data matrix is well conditioned:
        if np.linalg.cond(X) > MAX_CONDITION_CHECK:
            logger.error("Data is singular.")
            raise ValueError("Data matrix is nearly singular.")
        if self.channels is None:
            raise ValueError("Must define channels (set of basis functions).")
        if X.shape[0] < self.n_channels:
            logger.error("Not enough observations. Cannot calculate "
                         "pseudoinverse.")
            raise ValueError("Fewer observations (trials) than "
                             "channels. Cannot compute pseudoinverse.")
        # Check that the data matrix is the right size
        shape_data = np.shape(X)
        shape_labels = np.shape(y)
        if shape_data[0] != shape_labels[0]:
            raise ValueError("Mismatched data samples and label samples")
        if C is None:
            # Create a matrix of channel activations for every observation.
            # (i.e., C1 in Brouwer & Heeger 2009.)
            C = self._define_trial_activations(y)
        # Solve for W in B = WC
        self.W_ = X.transpose() @ np.linalg.pinv(C.transpose())
        if np.linalg.cond(self.W_) > MAX_CONDITION_CHECK:
            logger.error("Weight matrix is nearly singular.")
            raise ValueError("Weight matrix is nearly singular.")

        return self

    def predict(self, X):
        """Use test data to predict the feature

        Parameters
        ----------
        X: numpy matrix of voxel activation from test trials [observations,
            voxels]. Used to predict feature associated with the given
            observation.

        Returns
        -------
        model_prediction: numpy array of estimated feature values.
        """
        # Check that the data matrix is the right size
        shape_data = np.shape(X)
        if len(shape_data) != 2:
            raise ValueError("Data matrix has too many or too few "
                             "dimensions.")

        model_prediction = self._predict_features(X)

        return model_prediction

    def score(self, X, y):
        """Calculate error measure of prediction, assuming that the predicted
        feature is at the maximum of the reconstructed values.

        To score the reconstructions against expected features defined in the
        stimulus domain (i.e. in pixels), see score_against_reconstructed().

        Parameters
        ----------
        X: numpy matrix of voxel activation from new data
            [observations,voxels]
        y: numpy array of stimulus features. [observations, 2]

        Returns
        -------
        score_value: the error measurement between the actual
            feature and predicted features, [observations].
        """
        pred_features = self.predict(X)
        ssres = np.sum((pred_features - y) ** 2, axis=1)
        sstot = np.sum((y - np.mean(y)) ** 2, axis=1)
        score_value = 1 - (ssres / sstot)

        return score_value

    def score_against_reconstructed(self, X, y, metric="euclidean"):
        """Calculates a distance metric between reconstructed features in
        the 2D stimulus domain (i.e. reconstructions in pixels) given
        some observations X, and expected features y. Expected features must
        also be in the pixel stimulus domain.

        To score the reconstructions against the expected maxima, see score().

        Parameters
        ----------
        X: numpy matrix of voxel activation from new data
            [observations, voxels]
        y: numpy array of the expected stimulus reconstruction values [pixels,
            observations].
        metric: string specifying the distance metric, either "euclidean" or
            "cosine".

        Returns
        -------
        score_value: the error measurement between the reconstructed feature
            values as the expected values, [observations].
        """
        yhat = self.predict_feature_responses(X)
        if metric == "euclidean":
            score_value = euclidean_distances(y.T, yhat.T)
        elif metric == "cosine":
            score_value = cosine_distances(y.T, yhat.T)
        return score_value[0, :]

    def get_params(self, deep: bool = True):
        """Returns model parameters.

        Returns
        -------
        params: parameter of this object
        """
        return {"n_channels": self.n_channels, "channel_exp": self.channel_exp,
                "stim_fov": self.stim_fov, "stim_pixels": self.stim_pixels,
                "stim_radius_px": self.stim_radius_px, "xp": self.xp,
                "yp": self.yp, "channels": self.channels, "channel_limits":
                    self.channel_limits}

    def set_params(self, **parameters):
        """Sets model parameters after initialization.

        Parameters
        ----------
            parameters: structure with parameters and change values
        """
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        self._check_params()
        return self

    def _make_2d_cosine(self, x, y, x_center, y_center, s):
        """Defines a 2D exponentiated cosine (isometric, e.g. constant width
        in x & y) for use as a basis function. Function goes to zero at the
        given size constant s. That is, the function is given by if r <= s:
        f(r) = (0.5 + 0.5*cos(r*pi/s)))**channel_exp else:       0 where r is
        the Euclidean distance from the center of the function. This will
        yield a Gaussian-like function, centered at (x_center, y_center).

        Parameters
        ----------
        x: x-coordinates of the stimulus space, [npixels, 1] matrix
        y: y-coordinates of the stimulus space, [npixels, 1] matrix
        x_center: x-coordinate of basis function centers (sequence, nchannels)
        y_center: y-coordinate of basis function centers (sequence, nchannels)
        s: size constant of the 2D cosine function. This is the radius where
            the function is non-zero.

        Returns
        -------
        cos_functions: basis functions defined in the 2D stimulus space.
            returns a [nchannels, npixels] matrix.
        """
        cos_functions = np.zeros((len(x_center), len(x)))
        for i in range(len(x_center)):
            myr = np.sqrt((x - x_center[i]) ** 2 + (y - y_center[i]) ** 2). \
                squeeze()
            qq = (myr <= s) * 1
            zp = ((0.5 * (1 + np.cos(myr * np.pi / s))) ** self.channel_exp)
            cos_functions[i, :] = zp * qq
        return cos_functions

    def _2d_cosine_sz_to_fwhm(self, size_constant):
        fwhm = 2 * size_constant \
               * np.arccos((0.5 ** (1 / self.channel_exp) - 0.5) / 0.5) / np.pi
        return fwhm

    def _2d_cosine_fwhm_to_sz(self, fwhm):
        """For an exponentiated 2D cosine basis function, converts the
        full-width half-maximum (FWHM) of that function to the function's
        size constant. The size constant is the variable s in the function
        below:
            if r <= s:   f(r) = (0.5 + 0.5*cos(r*pi/s)))**channel_exp
            else:       0 where r is the Euclidean distance from the center of
        the function.

        Parameters
        ----------
        fwhm: a float indicating the full-width half-maximum in stimulus space

        Returns
        -------
        sz: the size constant of the exponentiated cosine
        """
        sz = (0.5 * np.pi * fwhm) / \
             (np.arccos((0.5 ** (1 / self.channel_exp) - 0.5) / 0.5))
        return sz

    def define_basis_functions_sqgrid(self, nchannels, channel_size=None):
        """Define basis functions (aka channels) arrange in a square grid.
        Sets the self.channels parameter.

        Parameters
        ----------
        nchannels: number of channels in the x (horizontal) direction
        channel_size: the desired full-width half-maximum (FWHM) of the
            channel, in stimulus space.

        Returns
        -------
        self.channels: defines channels, a [nchannels, npixels] matrix.
        channel_centers: numpy array of the centers of each channel, given as
            [nchannels x 2] matrix
        """
        # If given a single value for nchannels, expand to make a square
        if not isinstance(nchannels, list):
            nchannels = [nchannels, nchannels]
        chan_xcenters = np.linspace(self.channel_limits[0][0],
                                    self.channel_limits[0][1], nchannels[0])
        chan_ycenters = np.linspace(self.channel_limits[1][0],
                                    self.channel_limits[1][1], nchannels[1])
        cx, cy = np.meshgrid(chan_xcenters, chan_ycenters)
        cx = cx.reshape(-1, 1)
        cy = cy.reshape(-1, 1)
        if channel_size is None:
            # To get even coverage, setting the channel FWHM to ~1.1x-1.2x the
            # spacing between the channels might work. (See Sprague et al. 2013
            # Methods & Supplementary Figure 3 -- this is for cosine exp = 7,
            # your mileage may vary for other exponents!).
            channel_size = 1.2 * (chan_xcenters[1] - chan_xcenters[0])
        cos_width = self._2d_cosine_fwhm_to_sz(channel_size)
        # define exponentiated function
        self.channels = self._make_2d_cosine(self.xp.reshape(-1, 1),
                                             self.yp.reshape(-1, 1), cx, cy,
                                             cos_width)
        self.n_channels = self.channels.shape[0]

        return self.channels, np.hstack([cx, cy])

    def define_basis_functions_trigrid(self, grid_radius, channel_size=None):
        """Define basis functions (aka channels) arranged in a triangular grid.

        Returns
        -------
        self.channels: defines channels, [nchannels, npixels] matrix.
        channel_centers: numpy array of the centers of each channel
        """
        x_dist = np.diff(self.channel_limits[0]) / (grid_radius * 2)
        y_dist = x_dist * np.sqrt(3) * 0.5
        trigrid = np.zeros((0, 2))
        xbase = np.expand_dims(np.arange(self.channel_limits[0][0],
                                         self.channel_limits[0][1],
                                         x_dist.item()), 1)
        for yi, y in enumerate(np.arange(self.channel_limits[1][0],
                                         self.channel_limits[1][1],
                                         y_dist.item())):
            if (yi % 2) == 0:
                xx = xbase.copy()
                yy = np.ones((xx.size, 1)) * y
            else:
                xx = xbase.copy() + x_dist / 2
                yy = np.ones((xx.size, 1)) * y
            trigrid = np.vstack(
                (trigrid, np.hstack((xx, yy))))

        if channel_size is None:
            # To get even coverage, setting the channel FWHM to ~1.1x-1.2x the
            # spacing between the channels might work. (See Sprague et al. 2013
            # Methods & Supplementary Figure 3 -- this is for cosine exp = 7,
            # your mileage may vary for other exponents!).
            channel_size = 1.1 * x_dist
        cos_width = self._2d_cosine_fwhm_to_sz(channel_size)
        self.channels = self._make_2d_cosine(self.xp.reshape(-1, 1),
                                             self.yp.reshape(-1, 1),
                                             trigrid[:, 0],
                                             trigrid[:, 1], cos_width)
        self.n_channels = self.channels.shape[0]

        return self.channels, trigrid

    def _define_trial_activations(self, stim_centers, stim_radius=None):
        """Defines a numpy matrix of predicted channel responses for each
        trial/observation. Assumes that the presented stimulus is circular in
        the 2D stimulus space. This can effectively be a single circular
        pixel if stim_radius=0.5.

        Parameters
        -------
        stim_centers: numpy array of 2D stimulus features for each observation,
            expected dimensions are [observations, 2].
        stim_radius: scalar value or array-like specifying the radius of the
            circular stimulus for each observation, [observations]. While
            this can be read-out from the property self.stim_radius_px,
            here the user can specify it in case they are retraining the
            model with new observations.

        Returns
        -------
        C: numpy array of predicted channel responses [observations, pixels]
        """
        nstim = stim_centers.shape[0]
        if self.stim_radius_px is None:
            if stim_radius is None:
                raise ValueError("No defined stimulus radius. Please set.")
            else:
                self.stim_radius_px = stim_radius
        if not isinstance(self.stim_radius_px, np.ndarray) or not isinstance(
                self.stim_radius_px, list):
            self.stim_radius_px = np.ones(nstim) * self.stim_radius_px
        # Create a mask for every stimulus observation in the stimulus domain
        stimulus_mask = np.zeros((self.xp.size, nstim))
        for i in range(nstim):
            rad_vals = ((self.xp.reshape(-1, 1) - stim_centers[i, 0]) ** 2 +
                        (self.yp.reshape(-1, 1) - stim_centers[i, 1]) ** 2)
            inds = np.where(rad_vals < self.stim_radius_px[i])[0]
            stimulus_mask[inds, i] = 1
        # Go from the stimulus domain to the channel domain
        C = self.channels.squeeze() @ stimulus_mask
        C = C.transpose()
        # Check that C is full rank
        if np.linalg.matrix_rank(C) < self.n_channels:
            warnings.warn("Stimulus matrix is {}, not full rank. May cause "
                          "issues with stimulus prediction/reconstruction.".
                          format(np.linalg.matrix_rank(C)), RuntimeWarning)
        return C

    def _predict_channel_responses(self, X):
        """Computes predicted channel responses from data
        (e.g. C2 in Brouwer & Heeger 2009)

        Parameters
        ----------
        X: numpy data matrix. [observations, voxels]

        Returns
        -------
        channel_response: numpy matrix of channel responses. [channels,
            observations]
        """
        channel_response = np.matmul(np.linalg.pinv(self.W_), X.transpose())
        return channel_response

    def predict_feature_responses(self, X):
        """Takes channel weights and transforms them into continuous
        functions defined in the feature domain.

        Parameters
        ---------
        X: numpy matrix of data. [observations, voxels]

        Returns
        -------
        pred_response: predict response from all channels. This is the stimulus
            reconstruction in the channel domain. [pixels, observations]
        """
        pred_response = np.matmul(self.channels.transpose(),
                                  self._predict_channel_responses(X))
        return pred_response

    def _predict_features(self, X):
        """Predicts feature value from data in X.
        Takes the maximum of the reconstructed, i.e. predicted response
        function.

        Parameters
        ---------
        X: numpy matrix of data. [observations, voxels]

        Returns
        -------
        pred_features: numpy matrix of predicted stimulus features.
            [observations, 2]
        """
        pred_response = self.predict_feature_responses(X)
        feature_ind = np.argmax(pred_response, 0)
        pred_features = np.hstack((self.xp.reshape(-1, 1)[feature_ind],
                                   self.yp.reshape(-1, 1)[feature_ind]))

        return pred_features



import tensorflow as tf
import numpy as np
from sklearn.base import BaseEstimator
from brainiak.matnormal.matnormal_likelihoods import matnorm_logp
from brainiak.matnormal.utils import (
    pack_trainable_vars,
    unpack_trainable_vars,
    make_val_and_grad,
)
from scipy.optimize import minimize

__all__ = ["MatnormalRegression"]


class MatnormalRegression(BaseEstimator):
    """ This analysis allows maximum likelihood estimation of regression models
    in the presence of both spatial and temporal covariance.

    ..math::
        Y \\sim \\mathcal{MN}(X\beta, time_cov, space_cov)

    Parameters
    ----------
    time_cov : subclass of CovBase
        TR noise covariance class following CovBase interface.
    space_cov : subclass of CovBase
        Voxel noise covariance class following CovBase interface.
    optimizer : string, default="L-BFGS-B"
        Scipy optimizer to use. For other options, see "method" argument
        of scipy.optimize.minimize
    optCtrl: dict, default=None
        Additional arguments to pass to scipy.optimize.minimize.

    """

    def __init__(self, time_cov, space_cov, optimizer="L-BFGS-B",
                 optCtrl=None):

        self.optMethod = optimizer
        if optCtrl is None:
            self.optCtrl = {}

        self.time_cov = time_cov
        self.space_cov = space_cov

        self.n_t = time_cov.size
        self.n_v = space_cov.size

    def logp(self, X, Y):
        """ Log likelihood of model (internal)
        """
        y_hat = tf.matmul(X, self.beta)
        resid = Y - y_hat
        return matnorm_logp(resid, self.time_cov, self.space_cov)

    def fit(self, X, y, naive_init=True):
        """ Compute the regression fit.

        Parameters
        ----------
        X : np.array, TRs by conditions.
            Design matrix
        y : np.array, TRs by voxels.
            fMRI data
        """

        self.n_c = X.shape[1]

        if naive_init:
            # initialize to the least squares solution (basically all
            # we need now is the cov)
            sigma_inv_x = self.time_cov.solve(X)
            sigma_inv_y = self.time_cov.solve(y)

            beta_init = np.linalg.solve(
                (X.T).dot(sigma_inv_x), (X.T).dot(sigma_inv_y))

        else:
            beta_init = np.random.randn(self.n_c, self.n_v)

        self.beta = tf.Variable(beta_init, name="beta")

        self.train_variables = [self.beta]
        self.train_variables.extend(self.time_cov.get_optimize_vars())
        self.train_variables.extend(self.space_cov.get_optimize_vars())

        def lossfn(theta):
            return -self.logp(X, y)

        val_and_grad = make_val_and_grad(lossfn, self.train_variables)
        x0 = pack_trainable_vars(self.train_variables)

        opt_results = minimize(
            fun=val_and_grad, x0=x0, jac=True, method=self.optMethod,
            **self.optCtrl
        )

        unpacked_theta = unpack_trainable_vars(
            opt_results.x, self.train_variables)

        for var, val in zip(self.train_variables, unpacked_theta):
            var.assign(val)

        self.beta_ = self.beta.numpy()

    def predict(self, X):
        """ Predict fMRI signal from design matrix.

        Parameters
        ----------
        X : np.array, TRs by conditions.
            Design matrix

        """

        return X.dot(self.beta_)

    def calibrate(self, Y):
        """ Decode design matrix from fMRI dataset, based on a previously
        trained mapping. This method just does naive MLE:

        .. math::
            X = Y \\Sigma_s^{-1}B^T(B \\Sigma_s^{-1} B^T)^{-1}

        Parameters
        ----------
        Y : np.array, TRs by voxels.
            fMRI dataset
        """

        if Y.shape[1] <= self.n_c:
            raise RuntimeError(
                "More conditions than voxels! System is singular,\
                                cannot decode."
            )

        # Sigma_s^{-1} B'
        Sigma_s_btrp = self.space_cov.solve(tf.transpose(a=self.beta))
        # Y Sigma_s^{-1} B'
        Y_Sigma_Btrp = tf.matmul(Y, Sigma_s_btrp).numpy()
        # (B Sigma_s^{-1} B')^{-1}
        B_Sigma_Btrp = tf.matmul(self.beta, Sigma_s_btrp).numpy()

        X_test = np.linalg.solve(B_Sigma_Btrp.T, Y_Sigma_Btrp.T).T

        return X_test



#  Copyright 2017 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""I/O functionality."""

__all__ = [
    "load_boolean_mask",
    "load_images",
    "load_images_from_dir",
    "load_labels",
    "save_as_nifti_file",
]

from pathlib import Path
from typing import Optional, Callable, Iterable, List, Union

import logging
import nibabel as nib
import numpy as np

from nibabel.nifti1 import Nifti1Pair
from nibabel.spatialimages import SpatialImage

from .image import SingleConditionSpec

logger = logging.getLogger(__name__)


def load_images_from_dir(in_dir: Union[str, Path], suffix: str = "nii.gz",
                         ) -> Iterable[SpatialImage]:
    """Load images from directory.

    For efficiency, returns an iterator, not a sequence, so the results cannot
    be accessed by indexing.

    For every new iteration through the images, load_images_from_dir must be
    called again.

    Parameters
    ----------
    in_dir:
        Path to directory.
    suffix:
        Only load images with names that end like this.

    Yields
    ------
    SpatialImage
        Image.

    Examples
    --------
    >>> images = list(io.load_images_from_dir("face_scene", "bet.nii.gz"))
    """
    if isinstance(in_dir, str):
        in_dir = Path(in_dir)
    files = sorted(in_dir.glob("*" + suffix))
    for f in files:
        logger.debug(
            'Starting to read file %s', f
        )
        yield nib.load(str(f))  # type: ignore


def load_images(image_paths: Iterable[Union[str, Path]]
                ) -> Iterable[SpatialImage]:
    """Load images from paths.

    For efficiency, returns an iterator, not a sequence, so the results cannot
    be accessed by indexing.

    For every new iteration through the images, load_images must be called
    again.

    Parameters
    ----------
    image_paths:
        Paths to images.

    Yields
    ------
    SpatialImage
        Image.
    """
    for image_path in image_paths:
        if isinstance(image_path, Path):
            string_path = str(image_path)
        else:
            string_path = image_path
        logger.debug(
            'Starting to read file %s', string_path
        )
        yield nib.load(string_path)  # type: ignore


def load_boolean_mask(path: Union[str, Path],
                      predicate: Optional[
                          Callable[[np.ndarray], np.ndarray]] = None
                      ) -> np.ndarray:
    """Load boolean nibabel.SpatialImage mask.

    Parameters
    ----------
    path
        Mask path.
    predicate
        Callable used to create boolean values, e.g. a threshold function
        ``lambda x: x > 50``.

    Returns
    -------
    np.ndarray
        Boolean array corresponding to mask.
    """
    if not isinstance(path, str):
        path = str(path)
    data = nib.load(path).get_fdata()  # type: ignore
    if predicate is not None:
        mask = predicate(data)
    else:
        mask = data.astype(bool)
    return mask


def load_labels(path: Union[str, Path]) -> List[SingleConditionSpec]:
    """Load labels files.

    Parameters
    ----------
    path
        Path of labels file.

    Returns
    -------
    List[SingleConditionSpec]
        List of SingleConditionSpec stored in labels file.
    """
    condition_specs = np.load(str(path))
    return [c.view(SingleConditionSpec) for c in condition_specs]


def save_as_nifti_file(data: np.ndarray, affine: np.ndarray,
                       path: Union[str, Path]) -> None:
    """Create a Nifti file and save it.

    Parameters
    ----------
    data
        Brain data.
    affine
        Affine of the image, usually inherited from an existing image.
    path
        Output filename.
    """
    if not isinstance(path, str):
        path = str(path)
    img = Nifti1Pair(data, affine)
    nib.nifti1.save(img, path)
